"""Tests for active retrieval: contradiction detection and clarification callbacks."""

from __future__ import annotations

import json
import tempfile
from typing import List

import numpy as np
import pytest

from widemem.core.memory import WideMemory
from widemem.core.pipeline import AddResult
from widemem.core.types import (
    EmbeddingConfig,
    Fact,
    LLMConfig,
    Memory,
    MemoryConfig,
    MemorySearchResult,
    VectorStoreConfig,
)
from widemem.extraction.base import BaseExtractor
from widemem.providers.embeddings.base import BaseEmbedder
from widemem.providers.llm.base import BaseLLM
from widemem.retrieval.active import ActiveRetrieval, Clarification
from widemem.storage.vector.faiss_store import FAISSVectorStore


class MockLLM(BaseLLM):
    def __init__(self):
        super().__init__(LLMConfig())
        self.responses: list = []

    def set_responses(self, *responses):
        self.responses = list(responses)

    def generate(self, prompt, system=None):
        return json.dumps(self.generate_json(prompt, system))

    def generate_json(self, prompt, system=None):
        if self.responses:
            return self.responses.pop(0)
        return {}


class MockEmbedder(BaseEmbedder):
    def __init__(self, dimensions=64):
        super().__init__(EmbeddingConfig(dimensions=dimensions))

    def embed(self, text):
        rng = np.random.RandomState(hash(text) % 2**31)
        vec = rng.randn(self.config.dimensions).astype(np.float32)
        return (vec / np.linalg.norm(vec)).tolist()

    def embed_batch(self, texts):
        return [self.embed(t) for t in texts]


class DirectExtractor(BaseExtractor):
    def __init__(self):
        self.facts: List[Fact] = []

    def set_facts(self, *facts):
        self.facts = list(facts)

    def extract(self, text):
        if self.facts:
            return list(self.facts)
        return [Fact(content=text, importance=5.0)]


@pytest.fixture
def tmp_dir():
    with tempfile.TemporaryDirectory() as d:
        yield d


@pytest.fixture
def mock_llm():
    return MockLLM()


# --- ActiveRetrieval Unit Tests ---

class TestActiveRetrieval:
    def test_no_conflicts_when_no_existing(self, mock_llm):
        ar = ActiveRetrieval(mock_llm)
        facts = [Fact(content="Lives in Berlin")]
        result = ar.detect_conflicts(facts, [])
        assert result == []

    def test_no_conflicts_when_no_facts(self, mock_llm):
        ar = ActiveRetrieval(mock_llm)
        result = ar.detect_conflicts([], [
            MemorySearchResult(memory=Memory(content="old"), similarity_score=0.9)
        ])
        assert result == []

    def test_no_conflicts_below_similarity_threshold(self, mock_llm):
        ar = ActiveRetrieval(mock_llm, similarity_threshold=0.8)
        facts = [Fact(content="Lives in Berlin")]
        existing = [
            MemorySearchResult(memory=Memory(content="Likes coffee"), similarity_score=0.3)
        ]
        result = ar.detect_conflicts(facts, existing)
        assert result == []

    def test_detects_contradiction(self, mock_llm):
        ar = ActiveRetrieval(mock_llm, similarity_threshold=0.5)
        facts = [Fact(content="Lives in Berlin")]
        existing = [
            MemorySearchResult(
                memory=Memory(id="mem-1", content="Lives in Paris"),
                similarity_score=0.85,
            )
        ]

        mock_llm.set_responses({
            "has_conflict": True,
            "conflicts": [{
                "new_fact": "Lives in Berlin",
                "existing_memory_id": 1,
                "existing_content": "Lives in Paris",
                "type": "contradiction",
                "question": "Did you move from Paris to Berlin?",
            }]
        })

        result = ar.detect_conflicts(facts, existing)
        assert len(result) == 1
        assert result[0].conflict_type == "contradiction"
        assert result[0].question == "Did you move from Paris to Berlin?"
        assert result[0].existing_memory_id == "mem-1"

    def test_detects_ambiguity(self, mock_llm):
        ar = ActiveRetrieval(mock_llm, similarity_threshold=0.5)
        facts = [Fact(content="In Berlin this week")]
        existing = [
            MemorySearchResult(
                memory=Memory(id="mem-1", content="Lives in Paris"),
                similarity_score=0.7,
            )
        ]

        mock_llm.set_responses({
            "has_conflict": True,
            "conflicts": [{
                "new_fact": "In Berlin this week",
                "existing_memory_id": 1,
                "existing_content": "Lives in Paris",
                "type": "ambiguity",
                "question": "Are you visiting Berlin or did you move there?",
            }]
        })

        result = ar.detect_conflicts(facts, existing)
        assert len(result) == 1
        assert result[0].conflict_type == "ambiguity"

    def test_no_conflict_detected_by_llm(self, mock_llm):
        ar = ActiveRetrieval(mock_llm, similarity_threshold=0.5)
        facts = [Fact(content="Enjoys hiking")]
        existing = [
            MemorySearchResult(
                memory=Memory(content="Lives in Paris"),
                similarity_score=0.6,
            )
        ]

        mock_llm.set_responses({"has_conflict": False, "conflicts": []})

        result = ar.detect_conflicts(facts, existing)
        assert result == []

    def test_llm_error_gracefully_handled(self, mock_llm):
        ActiveRetrieval(mock_llm, similarity_threshold=0.5)

        # Make LLM raise
        class BrokenLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig())
            def generate(self, prompt, system=None):
                raise RuntimeError("LLM down")
            def generate_json(self, prompt, system=None):
                raise RuntimeError("LLM down")

        ar_broken = ActiveRetrieval(BrokenLLM(), similarity_threshold=0.5)
        facts = [Fact(content="test")]
        existing = [MemorySearchResult(memory=Memory(content="old"), similarity_score=0.9)]

        result = ar_broken.detect_conflicts(facts, existing)
        assert result == []


# --- Clarification Model Tests ---

class TestClarificationModel:
    def test_clarification_fields(self):
        c = Clarification(
            new_fact="Lives in Berlin",
            existing_content="Lives in Paris",
            existing_memory_id="mem-1",
            conflict_type="contradiction",
            question="Did you move?",
        )
        assert c.new_fact == "Lives in Berlin"
        assert c.conflict_type == "contradiction"
        assert c.question == "Did you move?"


# --- AddResult Tests ---

class TestAddResult:
    def test_no_clarifications(self):
        r = AddResult(memories=[Memory(content="test")])
        assert not r.has_clarifications
        assert r.clarifications == []

    def test_with_clarifications(self):
        c = Clarification(
            new_fact="a", existing_content="b",
            conflict_type="contradiction", question="which?",
        )
        r = AddResult(memories=[], clarifications=[c])
        assert r.has_clarifications
        assert len(r.clarifications) == 1


# --- Integration Tests ---

class TestActiveRetrievalIntegration:
    def _make_memory(self, tmp_dir, mock_llm, enable_active=True):
        emb = MockEmbedder(dimensions=64)
        vs = FAISSVectorStore(VectorStoreConfig(path=f"{tmp_dir}/vectors"), dimensions=64)
        config = MemoryConfig(
            history_db_path=f"{tmp_dir}/history.db",
            embedding=EmbeddingConfig(dimensions=64),
            enable_active_retrieval=enable_active,
            active_retrieval_threshold=-1.0,  # Accept all similarities in tests
        )
        mem = WideMemory(config=config, llm=mock_llm, embedder=emb, vector_store=vs)
        ext = DirectExtractor()
        mem.pipeline.extractor = ext
        return mem, ext

    def test_add_with_no_conflict(self, tmp_dir, mock_llm):
        mem, ext = self._make_memory(tmp_dir, mock_llm)

        ext.set_facts(Fact(content="Lives in Berlin", importance=7.0))
        result = mem.add("Lives in Berlin", user_id="alice")
        assert len(result.memories) == 1
        assert not result.has_clarifications

    def test_add_with_conflict_and_callback(self, tmp_dir, mock_llm):
        mem, ext = self._make_memory(tmp_dir, mock_llm)

        # First add
        ext.set_facts(Fact(content="Lives in Paris", importance=7.0))
        mem.add("Lives in Paris", user_id="alice")

        # Second add triggers conflict
        ext.set_facts(Fact(content="Lives in Berlin", importance=7.0))

        mock_llm.set_responses(
            # conflict detection
            {
                "has_conflict": True,
                "conflicts": [{
                    "new_fact": "Lives in Berlin",
                    "existing_memory_id": 1,
                    "existing_content": "Lives in Paris",
                    "type": "contradiction",
                    "question": "Did you move from Paris to Berlin?",
                }]
            },
            # conflict resolution (batch resolver)
            {"actions": [{"fact_index": 0, "action": "add", "target_id": None, "importance": 7}]},
        )

        callback_called = []

        def on_clarification(clarifications):
            callback_called.extend(clarifications)
            return ["Yes, I moved"]

        result = mem.add("Lives in Berlin", user_id="alice", on_clarification=on_clarification)
        assert result.has_clarifications
        assert len(callback_called) == 1
        assert callback_called[0].question == "Did you move from Paris to Berlin?"
        # Memories were still added because callback returned responses (not None)
        assert len(result.memories) >= 1

    def test_callback_returning_none_aborts_add(self, tmp_dir, mock_llm):
        mem, ext = self._make_memory(tmp_dir, mock_llm)

        # First add
        ext.set_facts(Fact(content="Lives in Paris", importance=7.0))
        mem.add("Lives in Paris", user_id="alice")

        # Second add with conflict
        ext.set_facts(Fact(content="Lives in Berlin", importance=7.0))

        mock_llm.set_responses({
            "has_conflict": True,
            "conflicts": [{
                "new_fact": "Lives in Berlin",
                "existing_memory_id": 1,
                "existing_content": "Lives in Paris",
                "type": "contradiction",
                "question": "Did you move?",
            }]
        })

        def abort_callback(clarifications):
            return None  # abort

        result = mem.add("Lives in Berlin", user_id="alice", on_clarification=abort_callback)
        assert result.has_clarifications
        assert len(result.memories) == 0  # nothing was added

    def test_active_retrieval_disabled(self, tmp_dir, mock_llm):
        mem, ext = self._make_memory(tmp_dir, mock_llm, enable_active=False)

        ext.set_facts(Fact(content="Lives in Paris", importance=7.0))
        mem.add("Lives in Paris", user_id="alice")

        ext.set_facts(Fact(content="Lives in Berlin", importance=7.0))
        result = mem.add("Lives in Berlin", user_id="alice")
        # No conflict detection when disabled
        assert not result.has_clarifications
        assert len(result.memories) >= 1

    def test_add_without_callback_still_returns_clarifications(self, tmp_dir, mock_llm):
        mem, ext = self._make_memory(tmp_dir, mock_llm)

        ext.set_facts(Fact(content="Lives in Paris", importance=7.0))
        mem.add("Lives in Paris", user_id="alice")

        ext.set_facts(Fact(content="Lives in Berlin", importance=7.0))

        mock_llm.set_responses(
            {
                "has_conflict": True,
                "conflicts": [{
                    "new_fact": "Lives in Berlin",
                    "existing_memory_id": 1,
                    "existing_content": "Lives in Paris",
                    "type": "contradiction",
                    "question": "Did you move?",
                }]
            },
            {"actions": [{"fact_index": 0, "action": "add", "target_id": None, "importance": 7}]},
        )

        # No callback — should still proceed and report clarifications
        result = mem.add("Lives in Berlin", user_id="alice")
        assert result.has_clarifications
        assert len(result.memories) >= 1



"""Basic widemem usage: add memories and search them."""

from widemem import WideMemory, MemoryConfig

# Default config uses OpenAI for LLM + embeddings, FAISS for storage
mem = WideMemory()

# Add memories
result = mem.add("I live in Berlin and work at Google as a senior engineer.", user_id="alice")
print(f"Added {len(result.memories)} memories")
for m in result.memories:
    print(f"  - {m.content} (importance: {m.importance})")

# Search
results = mem.search("where does alice work", user_id="alice", top_k=3)
for r in results:
    print(f"  [{r.final_score:.3f}] {r.memory.content}")

# --- Using Anthropic ---
# config = MemoryConfig(llm=LLMConfig(provider="anthropic", model="claude-sonnet-4-20250514"))
# mem = WideMemory(config=config)

# --- Fully local (Ollama + sentence-transformers) ---
# from widemem.core.types import LLMConfig, EmbeddingConfig
# config = MemoryConfig(
#     llm=LLMConfig(provider="ollama", model="llama3"),
#     embedding=EmbeddingConfig(provider="sentence-transformers", model="all-MiniLM-L6-v2", dimensions=384),
# )
# mem = WideMemory(config=config)



from __future__ import annotations

import json
import sqlite3
import uuid
from datetime import datetime
from pathlib import Path
from typing import List, Optional

from widemem.core.types import Fact


class ExtractionCollector:
    """Logs LLM extraction pairs (input text → extracted facts) as training data."""

    def __init__(self, db_path: str = "~/.widemem/extractions.db") -> None:
        path = Path(db_path).expanduser()
        path.parent.mkdir(parents=True, exist_ok=True)
        self.conn = sqlite3.connect(str(path))
        self._init_db()

    def _init_db(self) -> None:
        self.conn.execute("""
            CREATE TABLE IF NOT EXISTS extractions (
                id TEXT PRIMARY KEY,
                input_text TEXT NOT NULL,
                facts_json TEXT NOT NULL,
                model TEXT,
                timestamp TEXT NOT NULL
            )
        """)
        self.conn.commit()

    def log(
        self,
        input_text: str,
        facts: List[Fact],
        model: Optional[str] = None,
    ) -> str:
        entry_id = str(uuid.uuid4())
        facts_json = json.dumps([{"content": f.content, "importance": f.importance} for f in facts])
        self.conn.execute(
            "INSERT INTO extractions (id, input_text, facts_json, model, timestamp) VALUES (?, ?, ?, ?, ?)",
            (entry_id, input_text, facts_json, model, datetime.utcnow().isoformat()),
        )
        self.conn.commit()
        return entry_id

    def export(self, output_path: str, limit: Optional[int] = None) -> int:
        query = "SELECT input_text, facts_json FROM extractions ORDER BY timestamp"
        if limit:
            query += f" LIMIT {limit}"

        cursor = self.conn.execute(query)
        rows = cursor.fetchall()

        path = Path(output_path).expanduser()
        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "w") as f:
            for input_text, facts_json in rows:
                f.write(json.dumps({"input": input_text, "output": json.loads(facts_json)}) + "\n")

        return len(rows)

    def count(self) -> int:
        cursor = self.conn.execute("SELECT COUNT(*) FROM extractions")
        return cursor.fetchone()[0]

    def close(self) -> None:
        self.conn.close()



from __future__ import annotations

from typing import Dict, List, Optional

from widemem.core.types import Memory, MemoryAction, MemoryTier
from widemem.hierarchy.summarizer import MemorySummarizer
from widemem.providers.embeddings.base import BaseEmbedder
from widemem.storage.history import HistoryStore
from widemem.storage.vector.base import BaseVectorStore
from widemem.utils.hashing import content_hash


class HierarchyManager:
    def __init__(
        self,
        summarizer: MemorySummarizer,
        embedder: BaseEmbedder,
        vector_store: BaseVectorStore,
        history: HistoryStore,
        summarize_threshold: int = 10,
        theme_threshold: int = 3,
    ) -> None:
        self.summarizer = summarizer
        self.embedder = embedder
        self.vector_store = vector_store
        self.history = history
        self.summarize_threshold = summarize_threshold
        self.theme_threshold = theme_threshold

    def maybe_summarize(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        force: bool = False,
    ) -> List[Memory]:
        facts = self._get_memories_by_tier(MemoryTier.FACT, user_id=user_id, agent_id=agent_id)

        if not force and len(facts) < self.summarize_threshold:
            return []

        groups = self.summarizer.group_facts(facts)
        summaries = []

        for label, group_facts in groups:
            summary_text, importance = self.summarizer.summarize_group(group_facts)
            if not summary_text:
                continue

            memory = Memory(
                content=summary_text,
                user_id=user_id,
                agent_id=agent_id,
                tier=MemoryTier.SUMMARY,
                importance=importance,
                content_hash=content_hash(summary_text),
                metadata={"source_label": label, "source_count": len(group_facts)},
            )

            embedding = self.embedder.embed(summary_text)
            self.vector_store.insert(
                id=memory.id,
                vector=embedding,
                metadata=self._memory_to_metadata(memory),
            )
            self.history.log(memory.id, MemoryAction.ADD, new_content=summary_text)
            summaries.append(memory)

        if len(summaries) >= self.theme_threshold:
            theme_memories = self._synthesize_themes(summaries, user_id=user_id, agent_id=agent_id)
            summaries.extend(theme_memories)

        return summaries

    def _synthesize_themes(
        self,
        summaries: List[Memory],
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> List[Memory]:
        theme_text, importance = self.summarizer.synthesize_theme(summaries)
        if not theme_text:
            return []

        memory = Memory(
            content=theme_text,
            user_id=user_id,
            agent_id=agent_id,
            tier=MemoryTier.THEME,
            importance=importance,
            content_hash=content_hash(theme_text),
            metadata={"source_count": len(summaries)},
        )

        embedding = self.embedder.embed(theme_text)
        self.vector_store.insert(
            id=memory.id,
            vector=embedding,
            metadata=self._memory_to_metadata(memory),
        )
        self.history.log(memory.id, MemoryAction.ADD, new_content=theme_text)
        return [memory]

    def _get_memories_by_tier(
        self,
        tier: MemoryTier,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        max_results: int = 1000,
    ) -> List[Memory]:
        filters: Dict[str, str] = {"tier": tier.value}
        if user_id:
            filters["user_id"] = user_id
        if agent_id:
            filters["agent_id"] = agent_id

        results = self.vector_store.list_all(
            filters=filters,
            max_results=max_results,
        )

        return [
            Memory(
                id=id,
                content=metadata.get("content", ""),
                user_id=metadata.get("user_id"),
                agent_id=metadata.get("agent_id"),
                tier=MemoryTier(metadata.get("tier", "fact")),
                importance=metadata.get("importance", 5.0),
            )
            for id, metadata in results
        ]

    def _memory_to_metadata(self, memory: Memory) -> dict:
        meta = {
            "content": memory.content,
            "user_id": memory.user_id,
            "agent_id": memory.agent_id,
            "run_id": memory.run_id,
            "tier": memory.tier.value,
            "importance": memory.importance,
            "content_hash": memory.content_hash,
            "created_at": memory.created_at.isoformat(),
            "updated_at": memory.updated_at.isoformat(),
        }
        for k, v in memory.metadata.items():
            if k not in meta:
                meta[k] = v
        return meta



"""Hierarchical memory: facts -> summaries -> themes with query routing."""

from widemem import WideMemory, MemoryConfig

config = MemoryConfig(enable_hierarchy=True)
mem = WideMemory(config=config)

# Add many facts about a user
facts = [
    "I live in Berlin, Germany",
    "I moved from Paris last year",
    "I speak German and French fluently",
    "I work at Google as a senior engineer",
    "I specialize in distributed systems",
    "I enjoy hiking in the Alps on weekends",
    "I run marathons — my best time is 3:45",
    "I'm vegetarian and love cooking Thai food",
    "I have a golden retriever named Max",
    "I graduated from ETH Zurich in 2015",
]

for fact in facts:
    mem.add(fact, user_id="alice")

# Trigger summarization (groups facts, creates summaries and themes)
results = mem.summarize(user_id="alice", force=True)
print(f"Created {len(results)} summaries/themes:")
for m in results:
    print(f"  [{m.tier.value}] {m.content}")

# Specific query -> routed to facts
print("\n--- Specific query ---")
for r in mem.search("what is alice's marathon time", user_id="alice", top_k=3):
    print(f"  [{r.memory.tier.value}] {r.memory.content}")

# Broad query -> routed to themes/summaries
print("\n--- Broad query ---")
for r in mem.search("tell me about alice", user_id="alice", top_k=3):
    print(f"  [{r.memory.tier.value}] {r.memory.content}")



from __future__ import annotations

from abc import ABC, abstractmethod

from widemem.core.types import Fact


class BaseExtractor(ABC):
    @abstractmethod
    def extract(self, text: str) -> list[Fact]:
        ...



"""Tests for importance scoring, decay functions, and temporal retrieval."""

from __future__ import annotations

import math
from datetime import datetime, timedelta

import pytest

from widemem.core.types import (
    DecayFunction,
    Memory,
    MemorySearchResult,
    ScoringConfig,
)
from widemem.retrieval.temporal import score_and_rank
from widemem.scoring.decay import compute_recency_score
from widemem.scoring.importance import normalize_importance


class TestDecayFunctions:
    def test_no_decay(self):
        now = datetime(2026, 3, 8)
        old = datetime(2025, 1, 1)
        score = compute_recency_score(old, now, DecayFunction.NONE)
        assert score == 1.0

    def test_exponential_decay_recent(self):
        now = datetime(2026, 3, 8)
        yesterday = now - timedelta(days=1)
        score = compute_recency_score(yesterday, now, DecayFunction.EXPONENTIAL, decay_rate=0.01)
        assert score == pytest.approx(math.exp(-0.01), rel=1e-6)

    def test_exponential_decay_old(self):
        now = datetime(2026, 3, 8)
        old = now - timedelta(days=100)
        score = compute_recency_score(old, now, DecayFunction.EXPONENTIAL, decay_rate=0.01)
        assert score == pytest.approx(math.exp(-1.0), rel=1e-6)

    def test_exponential_decay_very_old(self):
        now = datetime(2026, 3, 8)
        old = now - timedelta(days=365)
        score = compute_recency_score(old, now, DecayFunction.EXPONENTIAL, decay_rate=0.01)
        assert score < 0.05

    def test_linear_decay(self):
        now = datetime(2026, 3, 8)
        ten_days_ago = now - timedelta(days=10)
        score = compute_recency_score(ten_days_ago, now, DecayFunction.LINEAR, decay_rate=0.01)
        assert score == pytest.approx(0.9, rel=1e-6)

    def test_linear_decay_floors_at_zero(self):
        now = datetime(2026, 3, 8)
        old = now - timedelta(days=200)
        score = compute_recency_score(old, now, DecayFunction.LINEAR, decay_rate=0.01)
        assert score == 0.0

    def test_step_decay_within_week(self):
        now = datetime(2026, 3, 8)
        recent = now - timedelta(days=3)
        assert compute_recency_score(recent, now, DecayFunction.STEP) == 1.0

    def test_step_decay_within_month(self):
        now = datetime(2026, 3, 8)
        two_weeks = now - timedelta(days=14)
        assert compute_recency_score(two_weeks, now, DecayFunction.STEP) == 0.7

    def test_step_decay_within_quarter(self):
        now = datetime(2026, 3, 8)
        two_months = now - timedelta(days=60)
        assert compute_recency_score(two_months, now, DecayFunction.STEP) == 0.4

    def test_step_decay_very_old(self):
        now = datetime(2026, 3, 8)
        old = now - timedelta(days=180)
        assert compute_recency_score(old, now, DecayFunction.STEP) == 0.1

    def test_same_time_returns_one(self):
        now = datetime(2026, 3, 8)
        for fn in [DecayFunction.EXPONENTIAL, DecayFunction.LINEAR, DecayFunction.STEP]:
            assert compute_recency_score(now, now, fn) == 1.0


class TestImportanceScoring:
    def test_normalize_mid(self):
        assert normalize_importance(5.0) == 0.5

    def test_normalize_max(self):
        assert normalize_importance(10.0) == 1.0

    def test_normalize_min(self):
        assert normalize_importance(0.0) == 0.0

    def test_normalize_clamps_high(self):
        assert normalize_importance(15.0) == 1.0

    def test_normalize_clamps_low(self):
        assert normalize_importance(-3.0) == 0.0


class TestScoreAndRank:
    def _make_result(self, content, similarity, importance, days_ago, now):
        return MemorySearchResult(
            memory=Memory(
                content=content,
                importance=importance,
                created_at=now - timedelta(days=days_ago),
            ),
            similarity_score=similarity,
        )

    def test_recent_high_importance_wins(self):
        now = datetime(2026, 3, 8)
        config = ScoringConfig()

        results = [
            self._make_result("old low", similarity=0.9, importance=3.0, days_ago=60, now=now),
            self._make_result("new high", similarity=0.8, importance=9.0, days_ago=1, now=now),
        ]

        ranked = score_and_rank(results, config, now=now)
        assert ranked[0].memory.content == "new high"

    def test_high_similarity_can_still_win(self):
        now = datetime(2026, 3, 8)
        config = ScoringConfig(similarity_weight=0.8, importance_weight=0.1, recency_weight=0.1)

        results = [
            self._make_result("exact match", similarity=0.99, importance=3.0, days_ago=30, now=now),
            self._make_result("vague match", similarity=0.3, importance=10.0, days_ago=0, now=now),
        ]

        ranked = score_and_rank(results, config, now=now)
        assert ranked[0].memory.content == "exact match"

    def test_time_filter_after(self):
        now = datetime(2026, 3, 8)
        config = ScoringConfig()
        cutoff = now - timedelta(days=7)

        results = [
            self._make_result("old", similarity=0.9, importance=8.0, days_ago=30, now=now),
            self._make_result("recent", similarity=0.8, importance=5.0, days_ago=3, now=now),
        ]

        ranked = score_and_rank(results, config, now=now, time_after=cutoff)
        assert len(ranked) == 1
        assert ranked[0].memory.content == "recent"

    def test_time_filter_before(self):
        now = datetime(2026, 3, 8)
        config = ScoringConfig()
        cutoff = now - timedelta(days=7)

        results = [
            self._make_result("old", similarity=0.9, importance=8.0, days_ago=30, now=now),
            self._make_result("recent", similarity=0.8, importance=5.0, days_ago=3, now=now),
        ]

        ranked = score_and_rank(results, config, now=now, time_before=cutoff)
        assert len(ranked) == 1
        assert ranked[0].memory.content == "old"

    def test_scores_are_populated(self):
        now = datetime(2026, 3, 8)
        config = ScoringConfig()

        results = [
            self._make_result("test", similarity=0.8, importance=7.0, days_ago=5, now=now),
        ]

        ranked = score_and_rank(results, config, now=now)
        r = ranked[0]
        assert r.similarity_score == 0.8
        assert r.importance_score == 0.7
        assert r.temporal_score > 0
        assert r.temporal_score <= 1.0
        assert r.final_score > 0

    def test_no_decay_config(self):
        now = datetime(2026, 3, 8)
        config = ScoringConfig(decay_function=DecayFunction.NONE)

        results = [
            self._make_result("old", similarity=0.5, importance=5.0, days_ago=365, now=now),
            self._make_result("new", similarity=0.5, importance=5.0, days_ago=1, now=now),
        ]

        ranked = score_and_rank(results, config, now=now)
        # With no decay, both should have the same final score
        assert ranked[0].final_score == pytest.approx(ranked[1].final_score, rel=1e-6)

    def test_empty_results(self):
        config = ScoringConfig()
        ranked = score_and_rank([], config)
        assert ranked == []


class TestTemporalIntegration:
    """Test that scoring integrates correctly with WideMemory.search()."""

    def test_search_returns_scored_results(self, tmp_dir):

        import numpy as np

        from widemem.core.memory import WideMemory
        from widemem.core.types import (
            EmbeddingConfig,
            Fact,
            LLMConfig,
            MemoryConfig,
            VectorStoreConfig,
        )
        from widemem.extraction.base import BaseExtractor
        from widemem.providers.embeddings.base import BaseEmbedder
        from widemem.providers.llm.base import BaseLLM

        class MockLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig())
            def generate(self, prompt, system=None):
                return "{}"
            def generate_json(self, prompt, system=None):
                return {"facts": []}

        class MockEmbedder(BaseEmbedder):
            def __init__(self):
                super().__init__(EmbeddingConfig(dimensions=64))
            def embed(self, text):
                rng = np.random.RandomState(hash(text) % 2**31)
                vec = rng.randn(64).astype(np.float32)
                return (vec / np.linalg.norm(vec)).tolist()
            def embed_batch(self, texts):
                return [self.embed(t) for t in texts]

        class DirectExtractor(BaseExtractor):
            def extract(self, text):
                return [Fact(content=text, importance=7.0)]

        from widemem.storage.vector.faiss_store import FAISSVectorStore
        config = MemoryConfig(
            history_db_path=f"{tmp_dir}/history.db",
            vector_store=VectorStoreConfig(path=f"{tmp_dir}/vectors"),
            embedding=EmbeddingConfig(dimensions=64),
        )
        embedder = MockEmbedder()
        vs = FAISSVectorStore(config.vector_store, dimensions=64)
        mem = WideMemory(config=config, llm=MockLLM(), embedder=embedder, vector_store=vs)
        mem.pipeline.extractor = DirectExtractor()

        mem.add("Lives in Berlin", user_id="alice")
        results = mem.search("Berlin", user_id="alice")

        assert len(results) >= 1
        r = results[0]
        assert r.final_score > 0
        assert r.temporal_score > 0
        assert r.importance_score > 0


@pytest.fixture
def tmp_dir():
    import tempfile
    with tempfile.TemporaryDirectory() as d:
        yield d



#!/usr/bin/env python3
"""End-to-end test using real OpenAI LLM + embeddings.

Requires OPENAI_API_KEY environment variable.

Usage:
    OPENAI_API_KEY=sk-... python scripts/e2e_test.py
"""

from __future__ import annotations

import os
import sys
import tempfile

def main():
    api_key = os.environ.get("OPENAI_API_KEY")
    if not api_key:
        print("ERROR: Set OPENAI_API_KEY environment variable")
        sys.exit(1)

    from widemem import WideMemory, MemoryConfig
    from widemem.core.types import (
        LLMConfig,
        EmbeddingConfig,
        VectorStoreConfig,
        YMYLConfig,
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        config = MemoryConfig(
            llm=LLMConfig(provider="openai", model="gpt-4o-mini", api_key=api_key),
            embedding=EmbeddingConfig(provider="openai", model="text-embedding-3-small", api_key=api_key, dimensions=1536),
            vector_store=VectorStoreConfig(provider="faiss", path=f"{tmpdir}/vectors"),
            history_db_path=f"{tmpdir}/history.db",
            ymyl=YMYLConfig(enabled=True),
        )

        with WideMemory(config) as memory:
            # --- Test 1: Basic add + search ---
            print("Test 1: Add and search...")
            result = memory.add(
                "My name is Alice. I live in Berlin and I work as a software engineer at Google.",
                user_id="alice",
            )
            assert len(result.memories) > 0, "Expected at least 1 memory"
            print(f"  Extracted {len(result.memories)} memories:")
            for m in result.memories:
                print(f"    - {m.content} (importance: {m.importance})")

            results = memory.search("where does alice live", user_id="alice", top_k=5)
            assert len(results) > 0, "Expected search results"
            print(f"  Search 'where does alice live': {len(results)} results")
            for r in results:
                print(f"    - {r.memory.content} (score: {r.final_score:.3f})")

            # --- Test 2: Update via contradiction ---
            print("\nTest 2: Contradiction handling...")
            result2 = memory.add(
                "Alice just moved to Paris. She no longer lives in Berlin.",
                user_id="alice",
            )
            print(f"  Extracted {len(result2.memories)} memories")
            for m in result2.memories:
                print(f"    - {m.content} (importance: {m.importance})")

            results2 = memory.search("where does alice live", user_id="alice", top_k=3)
            print(f"  Search after update: {len(results2)} results")
            for r in results2:
                print(f"    - {r.memory.content} (score: {r.final_score:.3f})")

            # --- Test 3: YMYL detection ---
            print("\nTest 3: YMYL content...")
            result3 = memory.add(
                "Alice has been diagnosed with diabetes. Her blood type is O+. She takes metformin daily.",
                user_id="alice",
            )
            print(f"  Extracted {len(result3.memories)} memories:")
            for m in result3.memories:
                print(f"    - {m.content} (importance: {m.importance})")

            high_importance = [m for m in result3.memories if m.importance >= 6.0]
            print(f"  High importance (>=6.0): {len(high_importance)} of {len(result3.memories)}")

            # --- Test 4: History ---
            print("\nTest 4: History audit trail...")
            if result.memories:
                history = memory.get_history(result.memories[0].id)
                print(f"  History for first memory: {len(history)} entries")
                for h in history:
                    print(f"    - {h.action.value}: {h.new_content or h.old_content}")

            # --- Test 5: Get and delete ---
            print("\nTest 5: Get and delete...")
            if result3.memories:
                mem_id = result3.memories[0].id
                got = memory.get(mem_id)
                assert got is not None, "Expected to get memory by id"
                print(f"  Get: {got.content}")

                memory.delete(mem_id)
                got_after = memory.get(mem_id)
                assert got_after is None, "Expected memory to be deleted"
                print("  Delete: confirmed deleted")

            # --- Test 6: Persistence ---
            print("\nTest 6: Persistence...")
            count_before = len(memory.search("alice", user_id="alice", top_k=50))

        with WideMemory(config) as memory2:
            count_after = len(memory2.search("alice", user_id="alice", top_k=50))
            print(f"  Before close: {count_before} results, after reload: {count_after} results")
            assert count_after > 0, "Expected memories to persist after reload"

    print("\n--- ALL E2E TESTS PASSED ---")


if __name__ == "__main__":
    main()



from __future__ import annotations

import math
from datetime import datetime

from widemem.core.types import DecayFunction


def compute_recency_score(
    created_at: datetime,
    now: datetime,
    decay_function: DecayFunction = DecayFunction.EXPONENTIAL,
    decay_rate: float = 0.01,
) -> float:
    age_days = max((now - created_at).total_seconds() / 86400, 0.0)

    if decay_function == DecayFunction.NONE:
        return 1.0

    if decay_function == DecayFunction.EXPONENTIAL:
        return math.exp(-decay_rate * age_days)

    if decay_function == DecayFunction.LINEAR:
        return max(1.0 - decay_rate * age_days, 0.0)

    if decay_function == DecayFunction.STEP:
        if age_days < 7:
            return 1.0
        if age_days < 30:
            return 0.7
        if age_days < 90:
            return 0.4
        return 0.1

    return 1.0



"""Integration tests for WideMemory using mock LLM and embedder."""

from __future__ import annotations

import json
import tempfile
from datetime import datetime

import numpy as np
import pytest

from widemem.conflict.batch_resolver import BatchConflictResolver
from widemem.core.memory import WideMemory
from widemem.core.types import (
    EmbeddingConfig,
    Fact,
    LLMConfig,
    MemoryConfig,
    MemorySearchResult,
    VectorStoreConfig,
)
from widemem.extraction.base import BaseExtractor
from widemem.providers.embeddings.base import BaseEmbedder
from widemem.providers.llm.base import BaseLLM
from widemem.storage.vector.faiss_store import FAISSVectorStore


class MockLLM(BaseLLM):
    def __init__(self) -> None:
        super().__init__(LLMConfig())
        self.responses: list[dict] = []

    def set_responses(self, *responses: dict) -> None:
        self.responses = list(responses)

    def _generate(self, prompt: str, system: str | None = None) -> str:
        return json.dumps(self._generate_json(prompt, system))

    def _generate_json(self, prompt: str, system: str | None = None) -> dict:
        if self.responses:
            return self.responses.pop(0)
        return {"facts": []}


class MockEmbedder(BaseEmbedder):
    def __init__(self, dimensions: int = 64) -> None:
        super().__init__(EmbeddingConfig(dimensions=dimensions))
        self._cache: dict[str, list[float]] = {}

    def embed(self, text: str) -> list[float]:
        if text not in self._cache:
            rng = np.random.RandomState(hash(text) % 2**31)
            vec = rng.randn(self.config.dimensions).astype(np.float32)
            vec = vec / np.linalg.norm(vec)
            self._cache[text] = vec.tolist()
        return self._cache[text]

    def embed_batch(self, texts: list[str]) -> list[list[float]]:
        return [self.embed(t) for t in texts]


class MockExtractor(BaseExtractor):
    def __init__(self) -> None:
        self.facts_to_return: list[Fact] = []

    def extract(self, text: str) -> list[Fact]:
        if self.facts_to_return:
            return self.facts_to_return
        return [Fact(content=text, importance=7.0)]


@pytest.fixture
def tmp_dir():
    with tempfile.TemporaryDirectory() as d:
        yield d


@pytest.fixture
def mock_llm():
    return MockLLM()


@pytest.fixture
def mock_embedder():
    return MockEmbedder(dimensions=64)


@pytest.fixture
def vector_store(tmp_dir):
    config = VectorStoreConfig(path=f"{tmp_dir}/vectors")
    return FAISSVectorStore(config, dimensions=64)


@pytest.fixture
def memory(tmp_dir, mock_llm, mock_embedder, vector_store):
    config = MemoryConfig(history_db_path=f"{tmp_dir}/history.db")
    mem = WideMemory(config=config, llm=mock_llm, embedder=mock_embedder, vector_store=vector_store)
    return mem


class TestFAISSVectorStore:
    def test_insert_and_search(self, vector_store, mock_embedder):
        vec = mock_embedder.embed("hello world")
        vector_store.insert("id1", vec, {"content": "hello world", "user_id": "alice"})

        results = vector_store.search(vec, top_k=1)
        assert len(results) == 1
        assert results[0][0] == "id1"
        assert results[0][2]["content"] == "hello world"

    def test_search_with_filters(self, vector_store, mock_embedder):
        vec1 = mock_embedder.embed("fact for alice")
        vec2 = mock_embedder.embed("fact for bob")
        vector_store.insert("id1", vec1, {"content": "fact for alice", "user_id": "alice"})
        vector_store.insert("id2", vec2, {"content": "fact for bob", "user_id": "bob"})

        results = vector_store.search(vec1, top_k=10, filters={"user_id": "alice"})
        assert all(r[2]["user_id"] == "alice" for r in results)

    def test_delete(self, vector_store, mock_embedder):
        vec = mock_embedder.embed("to delete")
        vector_store.insert("id1", vec, {"content": "to delete"})
        vector_store.delete("id1")
        result = vector_store.get("id1")
        assert result is None

    def test_update(self, vector_store, mock_embedder):
        vec1 = mock_embedder.embed("original")
        vector_store.insert("id1", vec1, {"content": "original"})

        vec2 = mock_embedder.embed("updated")
        vector_store.update("id1", vec2, {"content": "updated"})

        result = vector_store.get("id1")
        assert result is not None
        assert result[1]["content"] == "updated"

    def test_empty_search(self, vector_store, mock_embedder):
        vec = mock_embedder.embed("anything")
        results = vector_store.search(vec)
        assert results == []


class TestFAISSPersistence:
    def test_save_and_reload(self, tmp_dir):
        config = VectorStoreConfig(path=f"{tmp_dir}/persist_test")
        store = FAISSVectorStore(config, dimensions=4)

        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "hello"})
        store.insert("id2", [0.4, 0.3, 0.2, 0.1], {"content": "world"})

        store2 = FAISSVectorStore(config, dimensions=4)
        assert store2._index.ntotal == 2

        results = store2.search([0.1, 0.2, 0.3, 0.4], top_k=2)
        assert len(results) == 2
        contents = {r[2]["content"] for r in results}
        assert contents == {"hello", "world"}

    def test_get_after_reload(self, tmp_dir):
        config = VectorStoreConfig(path=f"{tmp_dir}/persist_get")
        store = FAISSVectorStore(config, dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "hello", "user_id": "alice"})

        store2 = FAISSVectorStore(config, dimensions=4)
        result = store2.get("id1")
        assert result is not None
        assert result[1]["content"] == "hello"
        assert result[1]["user_id"] == "alice"

    def test_insert_after_reload(self, tmp_dir):
        config = VectorStoreConfig(path=f"{tmp_dir}/persist_insert")
        store = FAISSVectorStore(config, dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "first"})

        store2 = FAISSVectorStore(config, dimensions=4)
        store2.insert("id2", [0.4, 0.3, 0.2, 0.1], {"content": "second"})
        assert store2._index.ntotal == 2

        store3 = FAISSVectorStore(config, dimensions=4)
        assert store3._index.ntotal == 2

    def test_delete_after_reload(self, tmp_dir):
        config = VectorStoreConfig(path=f"{tmp_dir}/persist_delete")
        store = FAISSVectorStore(config, dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "hello"})
        store.insert("id2", [0.4, 0.3, 0.2, 0.1], {"content": "world"})

        store2 = FAISSVectorStore(config, dimensions=4)
        store2.delete("id1")
        assert store2.get("id1") is None
        assert store2._index.ntotal == 1

        store3 = FAISSVectorStore(config, dimensions=4)
        assert store3.get("id1") is None
        assert store3._index.ntotal == 1

    def test_update_after_reload(self, tmp_dir):
        config = VectorStoreConfig(path=f"{tmp_dir}/persist_update")
        store = FAISSVectorStore(config, dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "v1"})

        store2 = FAISSVectorStore(config, dimensions=4)
        store2.update("id1", [0.5, 0.5, 0.0, 0.0], {"content": "v2"})

        store3 = FAISSVectorStore(config, dimensions=4)
        result = store3.get("id1")
        assert result is not None
        assert result[1]["content"] == "v2"
        assert store3._index.ntotal == 1

    def test_filters_after_reload(self, tmp_dir):
        config = VectorStoreConfig(path=f"{tmp_dir}/persist_filters")
        store = FAISSVectorStore(config, dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "a", "user_id": "alice"})
        store.insert("id2", [0.4, 0.3, 0.2, 0.1], {"content": "b", "user_id": "bob"})

        store2 = FAISSVectorStore(config, dimensions=4)
        results = store2.search([0.1, 0.2, 0.3, 0.4], top_k=10, filters={"user_id": "alice"})
        assert len(results) == 1
        assert results[0][2]["user_id"] == "alice"

    def test_no_path_means_no_persistence(self, tmp_dir):
        config = VectorStoreConfig(path=None)
        store = FAISSVectorStore(config, dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "ephemeral"})
        assert store._storage_path is None


class TestBatchConflictResolver:
    def test_all_add_when_no_existing(self, mock_llm):
        resolver = BatchConflictResolver(mock_llm)
        facts = [Fact(content="Lives in Berlin", importance=8.0)]
        actions = resolver.resolve(facts, [])

        assert len(actions) == 1
        assert actions[0].action.value == "add"
        assert actions[0].fact == "Lives in Berlin"

    def test_update_existing(self, mock_llm):
        resolver = BatchConflictResolver(mock_llm)
        facts = [Fact(content="Lives in Berlin", importance=8.0)]

        from widemem.core.types import Memory
        existing = [MemorySearchResult(
            memory=Memory(id="mem-1", content="Lives in Paris"),
            similarity_score=0.85,
        )]

        mock_llm.set_responses({
            "actions": [{"fact_index": 0, "action": "update", "target_id": 1, "importance": 8}]
        })

        actions = resolver.resolve(facts, existing)
        assert len(actions) == 1
        assert actions[0].action.value == "update"
        assert actions[0].target_id == "mem-1"


class TestWideMemory:
    def test_add_and_search(self, memory, mock_embedder):
        extractor = MockExtractor()
        extractor.facts_to_return = [Fact(content="Lives in Berlin", importance=8.0)]
        memory.pipeline.extractor = extractor

        result = memory.add("I live in Berlin", user_id="alice")
        assert len(result.memories) == 1
        assert result.memories[0].content == "Lives in Berlin"
        assert result.memories[0].user_id == "alice"

        search_results = memory.search("where does alice live", user_id="alice")
        assert len(search_results) >= 1
        assert search_results[0].memory.content == "Lives in Berlin"

    def test_add_multiple_facts(self, memory):
        extractor = MockExtractor()
        extractor.facts_to_return = [
            Fact(content="Lives in Berlin", importance=8.0),
            Fact(content="Works at Google", importance=7.0),
        ]
        memory.pipeline.extractor = extractor

        result = memory.add("I live in Berlin and work at Google", user_id="alice")
        assert len(result.memories) == 2

    def test_get_memory(self, memory):
        extractor = MockExtractor()
        extractor.facts_to_return = [Fact(content="Test fact", importance=5.0)]
        memory.pipeline.extractor = extractor

        result = memory.add("Test fact", user_id="alice")
        assert len(result.memories) == 1

        retrieved = memory.get(result.memories[0].id)
        assert retrieved is not None
        assert retrieved.content == "Test fact"

    def test_delete_memory(self, memory):
        extractor = MockExtractor()
        extractor.facts_to_return = [Fact(content="To be deleted", importance=5.0)]
        memory.pipeline.extractor = extractor

        result = memory.add("To be deleted", user_id="alice")
        memory.delete(result.memories[0].id)

        retrieved = memory.get(result.memories[0].id)
        assert retrieved is None


class TestBatchAdd:
    def test_add_batch_multiple_texts(self, memory):
        extractor = MockExtractor()
        memory.pipeline.extractor = extractor

        results = memory.add_batch(
            ["I live in Berlin", "I work at Google"],
            user_id="alice",
        )
        assert len(results) == 2
        assert all(len(r.memories) >= 1 for r in results)

    def test_add_batch_empty(self, memory):
        results = memory.add_batch([], user_id="alice")
        assert results == []


class TestMemoryCount:
    def test_count_all(self, memory):
        extractor = MockExtractor()
        memory.pipeline.extractor = extractor

        memory.add("fact one", user_id="alice")
        memory.add("fact two", user_id="alice")
        assert memory.count(user_id="alice") == 2

    def test_count_empty(self, memory):
        assert memory.count(user_id="alice") == 0

    def test_count_filtered(self, memory):
        extractor = MockExtractor()
        memory.pipeline.extractor = extractor

        memory.add("fact one", user_id="alice")
        memory.add("fact two", user_id="bob")
        assert memory.count(user_id="alice") == 1
        assert memory.count(user_id="bob") == 1


class TestExportImport:
    def test_export_json(self, memory):
        extractor = MockExtractor()
        memory.pipeline.extractor = extractor

        memory.add("Lives in Berlin", user_id="alice")
        data = memory.export_json(user_id="alice")
        parsed = json.loads(data)
        assert parsed["count"] == 1
        assert parsed["memories"][0]["content"] == "Lives in Berlin"

    def test_import_json(self, tmp_dir):
        config = MemoryConfig(history_db_path=f"{tmp_dir}/h2.db")
        embedder = MockEmbedder(dimensions=64)
        vs = FAISSVectorStore(VectorStoreConfig(), dimensions=64)
        mem = WideMemory(config=config, llm=MockLLM(), embedder=embedder, vector_store=vs)

        data = json.dumps({"memories": [
            {"id": "test-id-1", "content": "Lives in Berlin", "user_id": "alice", "importance": 8.0},
            {"id": "test-id-2", "content": "Works at Google", "user_id": "alice", "importance": 7.0},
        ]})
        imported = mem.import_json(data)
        assert imported == 2
        assert mem.count(user_id="alice") == 2

    def test_import_skips_existing(self, tmp_dir):
        config = MemoryConfig(history_db_path=f"{tmp_dir}/h3.db")
        embedder = MockEmbedder(dimensions=64)
        vs = FAISSVectorStore(VectorStoreConfig(), dimensions=64)
        mem = WideMemory(config=config, llm=MockLLM(), embedder=embedder, vector_store=vs)

        data = json.dumps({"memories": [
            {"id": "test-id-1", "content": "Lives in Berlin", "importance": 8.0},
        ]})
        assert mem.import_json(data) == 1
        assert mem.import_json(data) == 0

    def test_roundtrip(self, memory):
        extractor = MockExtractor()
        memory.pipeline.extractor = extractor
        memory.add("Lives in Berlin", user_id="alice")
        memory.add("Works at Google", user_id="alice")

        exported = memory.export_json(user_id="alice")
        parsed = json.loads(exported)
        assert parsed["count"] == 2


class TestTTL:
    def test_ttl_filters_old_memories(self, tmp_dir):
        config = MemoryConfig(history_db_path=f"{tmp_dir}/ttl.db", ttl_days=7)
        embedder = MockEmbedder(dimensions=64)
        vs = FAISSVectorStore(VectorStoreConfig(), dimensions=64)
        mem = WideMemory(config=config, llm=MockLLM(), embedder=embedder, vector_store=vs)

        from datetime import timedelta
        old_time = (datetime.utcnow() - timedelta(days=30)).isoformat()
        new_time = datetime.utcnow().isoformat()

        vec1 = embedder.embed("old fact")
        vs.insert("id1", vec1, {
            "content": "old fact", "created_at": old_time, "updated_at": old_time,
            "user_id": "alice", "tier": "fact", "importance": 5.0,
        })
        vec2 = embedder.embed("new fact")
        vs.insert("id2", vec2, {
            "content": "new fact", "created_at": new_time, "updated_at": new_time,
            "user_id": "alice", "tier": "fact", "importance": 5.0,
        })

        results = mem.search("fact", user_id="alice")
        contents = [r.memory.content for r in results]
        assert "new fact" in contents
        assert "old fact" not in contents

    def test_no_ttl_returns_all(self, tmp_dir):
        config = MemoryConfig(history_db_path=f"{tmp_dir}/nottl.db", ttl_days=None)
        embedder = MockEmbedder(dimensions=64)
        vs = FAISSVectorStore(VectorStoreConfig(), dimensions=64)
        mem = WideMemory(config=config, llm=MockLLM(), embedder=embedder, vector_store=vs)

        from datetime import timedelta
        old_time = (datetime.utcnow() - timedelta(days=365)).isoformat()

        vec = embedder.embed("ancient fact")
        vs.insert("id1", vec, {
            "content": "ancient fact", "created_at": old_time, "updated_at": old_time,
            "user_id": "alice", "tier": "fact", "importance": 5.0,
        })

        results = mem.search("ancient", user_id="alice")
        assert len(results) == 1


class TestRetryBackoff:
    def test_retry_on_transient_error(self):
        call_count = 0

        class FlakeyLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig(), max_retries=3, retry_delay=0.01)

            def _generate(self, prompt, system=None):
                nonlocal call_count
                call_count += 1
                if call_count < 3:
                    raise ConnectionError("network flake")
                return "ok"

            def _generate_json(self, prompt, system=None):
                return {}

        llm = FlakeyLLM()
        result = llm.generate("test")
        assert result == "ok"
        assert call_count == 3

    def test_provider_error_not_retried(self):
        from widemem.core.exceptions import ProviderError

        class BadLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig(), max_retries=3, retry_delay=0.01)

            def _generate(self, prompt, system=None):
                raise ProviderError("bad request")

            def _generate_json(self, prompt, system=None):
                return {}

        llm = BadLLM()
        with pytest.raises(ProviderError, match="bad request"):
            llm.generate("test")


class TestScoreBreakdown:
    def test_search_results_have_score_components(self, memory):
        extractor = MockExtractor()
        extractor.facts_to_return = [Fact(content="Lives in Berlin", importance=8.0)]
        memory.pipeline.extractor = extractor
        memory.add("I live in Berlin", user_id="alice")

        results = memory.search("Berlin", user_id="alice")
        assert len(results) >= 1
        r = results[0]
        assert r.similarity_score > 0
        assert r.temporal_score > 0
        assert r.importance_score > 0
        assert r.final_score > 0


class TestIDMapping:
    def test_uuid_to_int_mapping(self):
        from widemem.utils.id_mapping import IDMapper

        mapper = IDMapper()
        idx1 = mapper.add("uuid-1")
        idx2 = mapper.add("uuid-2")

        assert idx1 == 1
        assert idx2 == 2
        assert mapper.to_uuid(1) == "uuid-1"
        assert mapper.to_uuid(2) == "uuid-2"
        assert mapper.to_int("uuid-1") == 1

    def test_duplicate_add(self):
        from widemem.utils.id_mapping import IDMapper

        mapper = IDMapper()
        idx1 = mapper.add("uuid-1")
        idx2 = mapper.add("uuid-1")
        assert idx1 == idx2


class TestContentHash:
    def test_same_content_same_hash(self):
        from widemem.utils.hashing import content_hash

        assert content_hash("hello") == content_hash("hello")

    def test_different_content_different_hash(self):
        from widemem.utils.hashing import content_hash

        assert content_hash("hello") != content_hash("world")



from __future__ import annotations

import uuid
from typing import Any, Dict, List, Optional, Tuple

from widemem.core.exceptions import StorageError
from widemem.core.types import VectorStoreConfig
from widemem.storage.vector.base import BaseVectorStore


class QdrantVectorStore(BaseVectorStore):
    def __init__(
        self,
        config: VectorStoreConfig,
        dimensions: int = 1536,
        collection_name: str = "widemem",
    ) -> None:
        super().__init__(config)
        self.dimensions = dimensions
        self.collection_name = collection_name

        try:
            from qdrant_client import QdrantClient
            from qdrant_client.models import Distance, VectorParams
        except ImportError:
            raise StorageError("Install qdrant: pip install widemem[qdrant]")

        if config.path:
            self.client = QdrantClient(path=config.path)
        else:
            self.client = QdrantClient(url="localhost", port=6333)

        collections = [c.name for c in self.client.get_collections().collections]
        if collection_name not in collections:
            self.client.create_collection(
                collection_name=collection_name,
                vectors_config=VectorParams(
                    size=dimensions,
                    distance=Distance.COSINE,
                ),
            )

    def insert(self, id: str, vector: List[float], metadata: Dict[str, Any]) -> None:
        from qdrant_client.models import PointStruct

        self.client.upsert(
            collection_name=self.collection_name,
            points=[PointStruct(
                id=self._to_qdrant_id(id),
                vector=vector,
                payload={**metadata, "_widemem_id": id},
            )],
        )

    def search(
        self,
        vector: List[float],
        top_k: int = 10,
        filters: Optional[Dict[str, Any]] = None,
    ) -> List[Tuple[str, float, Dict[str, Any]]]:
        query_filter = None
        if filters:
            from qdrant_client.models import FieldCondition, Filter, MatchValue
            conditions = [
                FieldCondition(key=k, match=MatchValue(value=v))
                for k, v in filters.items()
            ]
            query_filter = Filter(must=conditions)

        results = self.client.query_points(
            collection_name=self.collection_name,
            query=vector,
            limit=top_k,
            query_filter=query_filter,
            with_payload=True,
        )

        output = []
        for point in results.points:
            payload = point.payload or {}
            widemem_id = payload.pop("_widemem_id", str(point.id))
            output.append((widemem_id, point.score, payload))
        return output

    def update(self, id: str, vector: List[float], metadata: Dict[str, Any]) -> None:
        self.insert(id, vector, metadata)

    def delete(self, id: str) -> None:
        from qdrant_client.models import PointIdsList
        self.client.delete(
            collection_name=self.collection_name,
            points_selector=PointIdsList(points=[self._to_qdrant_id(id)]),
        )

    def get(self, id: str) -> Optional[Tuple[List[float], Dict[str, Any]]]:
        results = self.client.retrieve(
            collection_name=self.collection_name,
            ids=[self._to_qdrant_id(id)],
            with_vectors=True,
            with_payload=True,
        )
        if not results:
            return None
        point = results[0]
        payload = point.payload or {}
        payload.pop("_widemem_id", None)
        vector = point.vector if isinstance(point.vector, list) else []
        return vector, payload

    def list_all(
        self,
        filters: Optional[Dict[str, Any]] = None,
        max_results: int = 1000,
    ) -> List[Tuple[str, Dict[str, Any]]]:
        query_filter = None
        if filters:
            from qdrant_client.models import FieldCondition, Filter, MatchValue
            conditions = [
                FieldCondition(key=k, match=MatchValue(value=v))
                for k, v in filters.items()
            ]
            query_filter = Filter(must=conditions)

        results = self.client.scroll(
            collection_name=self.collection_name,
            scroll_filter=query_filter,
            limit=max_results,
            with_payload=True,
            with_vectors=False,
        )

        output = []
        for point in results[0]:
            payload = point.payload or {}
            widemem_id = payload.pop("_widemem_id", str(point.id))
            output.append((widemem_id, payload))
        return output

    def _to_qdrant_id(self, id: str) -> str:
        try:
            uuid.UUID(id)
            return id
        except ValueError:
            return str(uuid.uuid5(uuid.NAMESPACE_DNS, id))



from __future__ import annotations

import json

from widemem.core.exceptions import ProviderError
from widemem.core.types import LLMConfig
from widemem.providers.llm.base import BaseLLM


class OllamaLLM(BaseLLM):
    def __init__(self, config: LLMConfig) -> None:
        super().__init__(config)
        try:
            from ollama import Client
        except ImportError:
            raise ProviderError("Install ollama: pip install widemem[ollama]")
        self.client = Client(host=config.base_url or "http://localhost:11434")

    def _generate(self, prompt: str, system: str | None = None) -> str:
        messages = []
        if system:
            messages.append({"role": "system", "content": system})
        messages.append({"role": "user", "content": prompt})

        response = self.client.chat(
            model=self.config.model,
            messages=messages,
            options={"temperature": self.config.temperature},
        )
        content = response.get("message", {}).get("content", "")
        if not content:
            raise ProviderError("Ollama returned empty response")
        return content

    def _generate_json(self, prompt: str, system: str | None = None) -> dict:
        json_system = (system or "") + "\n\nYou must respond with valid JSON only. No other text."
        text = self._generate(prompt, system=json_system.strip())

        text = text.strip()
        if text.startswith("```json"):
            text = text[7:]
        if text.startswith("```"):
            text = text[3:]
        if text.endswith("```"):
            text = text[:-3]
        text = text.strip()

        try:
            return json.loads(text)
        except json.JSONDecodeError as e:
            raise ProviderError(f"Ollama returned invalid JSON: {e}") from e



from __future__ import annotations

from widemem.conflict.prompts import (
    BATCH_CONFLICT_RESOLUTION_PROMPT,
    BATCH_CONFLICT_RESOLUTION_SYSTEM,
)
from widemem.core.types import ActionItem, Fact, MemoryAction, MemorySearchResult
from widemem.providers.llm.base import BaseLLM
from widemem.utils.id_mapping import IDMapper


class BatchConflictResolver:
    def __init__(self, llm: BaseLLM) -> None:
        self.llm = llm

    def resolve(
        self,
        new_facts: list[Fact],
        existing_memories: list[MemorySearchResult],
    ) -> list[ActionItem]:
        if not new_facts:
            return []

        if not existing_memories:
            return [
                ActionItem(action=MemoryAction.ADD, fact=f.content, importance=f.importance)
                for f in new_facts
            ]

        mapper = IDMapper()
        for mem_result in existing_memories:
            mapper.add(mem_result.memory.id)

        new_facts_str = "\n".join(
            f"[{i}] {f.content} (importance: {f.importance})"
            for i, f in enumerate(new_facts)
        )
        existing_str = "\n".join(
            f"[{mapper.to_int(m.memory.id)}] {m.memory.content}"
            for m in existing_memories
        )

        prompt = BATCH_CONFLICT_RESOLUTION_PROMPT.format(
            new_facts=new_facts_str,
            existing_memories=existing_str,
        )

        try:
            result = self.llm.generate_json(prompt, system=BATCH_CONFLICT_RESOLUTION_SYSTEM)
        except Exception:
            return [
                ActionItem(action=MemoryAction.ADD, fact=f.content, importance=f.importance)
                for f in new_facts
            ]

        actions = []
        seen_indices: set[int] = set()
        for item in result.get("actions", []):
            fact_idx = item.get("fact_index")
            if fact_idx is None or not isinstance(fact_idx, int) or fact_idx < 0 or fact_idx >= len(new_facts):
                continue
            if fact_idx in seen_indices:
                continue
            seen_indices.add(fact_idx)

            target_id = None
            raw_target = item.get("target_id")
            if raw_target is not None:
                try:
                    target_id = mapper.to_uuid(int(raw_target))
                except (ValueError, TypeError):
                    target_id = None

            action_str = item.get("action", "add").lower()
            try:
                action = MemoryAction(action_str)
            except ValueError:
                action = MemoryAction.ADD

            if action in (MemoryAction.UPDATE, MemoryAction.DELETE) and target_id is None:
                action = MemoryAction.ADD
            actions.append(ActionItem(
                action=action,
                fact=new_facts[fact_idx].content,
                target_id=target_id,
                importance=float(item.get("importance", new_facts[fact_idx].importance)),
            ))

        for i, fact in enumerate(new_facts):
            if i not in seen_indices:
                actions.append(ActionItem(
                    action=MemoryAction.ADD,
                    fact=fact.content,
                    importance=fact.importance,
                ))

        return actions



from __future__ import annotations

import json
from datetime import datetime, timedelta
from typing import Any, Callable, Dict, List, Optional

from widemem.conflict.batch_resolver import BatchConflictResolver
from widemem.core.pipeline import AddResult, MemoryPipeline
from widemem.core.types import (
    HistoryEntry,
    Memory,
    MemoryConfig,
    MemorySearchResult,
    MemoryTier,
)
from widemem.extraction.collector import ExtractionCollector
from widemem.extraction.llm_extractor import LLMExtractor
from widemem.hierarchy.manager import HierarchyManager
from widemem.hierarchy.query_router import classify_query, route_results
from widemem.hierarchy.summarizer import MemorySummarizer
from widemem.providers.embeddings.base import BaseEmbedder
from widemem.providers.embeddings.openai import OpenAIEmbedder
from widemem.providers.llm.base import BaseLLM
from widemem.providers.llm.openai import OpenAILLM
from widemem.retrieval.active import ActiveRetrieval, Clarification
from widemem.retrieval.temporal import score_and_rank
from widemem.storage.history import HistoryStore
from widemem.storage.vector.base import BaseVectorStore
from widemem.storage.vector.faiss_store import FAISSVectorStore


class WideMemory:
    def __init__(
        self,
        config: Optional[MemoryConfig] = None,
        llm: Optional[BaseLLM] = None,
        embedder: Optional[BaseEmbedder] = None,
        vector_store: Optional[BaseVectorStore] = None,
    ) -> None:
        self.config = config or MemoryConfig()

        self.llm = llm or self._create_llm()
        self.embedder = embedder or self._create_embedder()
        self.vector_store = vector_store or self._create_vector_store()
        self._history_store = HistoryStore(self.config.history_db_path)

        collector = (
            ExtractionCollector(self.config.extractions_db_path)
            if self.config.collect_extractions else None
        )
        extractor = LLMExtractor(
            self.llm,
            collector=collector,
            ymyl_config=self.config.ymyl,
            custom_topics=self.config.topics.custom_topics,
        )
        resolver = BatchConflictResolver(self.llm)
        active = (
            ActiveRetrieval(self.llm, similarity_threshold=self.config.active_retrieval_threshold)
            if self.config.enable_active_retrieval else None
        )

        ymyl_active = None
        if not active and self.config.ymyl.enabled and self.config.ymyl.force_active_retrieval:
            ymyl_active = ActiveRetrieval(
                self.llm, similarity_threshold=self.config.active_retrieval_threshold
            )

        self.pipeline = MemoryPipeline(
            extractor=extractor,
            resolver=resolver,
            embedder=self.embedder,
            vector_store=self.vector_store,
            history=self._history_store,
            active_retrieval=active,
            ymyl_active_retrieval=ymyl_active,
            ymyl_config=self.config.ymyl,
        )

        self._hierarchy = HierarchyManager(
            summarizer=MemorySummarizer(self.llm),
            embedder=self.embedder,
            vector_store=self.vector_store,
            history=self._history_store,
        )

    def close(self) -> None:
        self._history_store.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def add(
        self,
        text: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        on_clarification: Optional[Callable[[List[Clarification]], Optional[List[str]]]] = None,
    ) -> AddResult:
        return self.pipeline.process(
            text=text,
            user_id=user_id,
            agent_id=agent_id,
            run_id=run_id,
            on_clarification=on_clarification,
        )

    def add_batch(
        self,
        texts: List[str],
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
    ) -> List[AddResult]:
        results = []
        for text in texts:
            result = self.add(text, user_id=user_id, agent_id=agent_id, run_id=run_id)
            results.append(result)
        return results

    def search(
        self,
        query: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        top_k: int = 10,
        time_after: Optional[datetime] = None,
        time_before: Optional[datetime] = None,
        tier: Optional[MemoryTier] = None,
    ) -> List[MemorySearchResult]:
        embedding = self.embedder.embed(query)

        filters: Dict[str, Any] = {}
        if user_id:
            filters["user_id"] = user_id
        if agent_id:
            filters["agent_id"] = agent_id
        if tier:
            filters["tier"] = tier.value

        top_k = min(top_k, 1000)
        fetch_k = top_k * 3
        raw_results = self.vector_store.search(
            vector=embedding,
            top_k=fetch_k,
            filters=filters or None,
        )

        now = datetime.utcnow()
        ttl_cutoff = now - timedelta(days=self.config.ttl_days) if self.config.ttl_days else None

        search_results = []
        for id, score, metadata in raw_results:
            created_at = (
                datetime.fromisoformat(metadata["created_at"])
                if "created_at" in metadata else now
            )
            if ttl_cutoff and created_at < ttl_cutoff:
                continue
            search_results.append(MemorySearchResult(
                memory=Memory(
                    id=id,
                    content=metadata.get("content", ""),
                    user_id=metadata.get("user_id"),
                    agent_id=metadata.get("agent_id"),
                    importance=metadata.get("importance", 5.0),
                    tier=MemoryTier(metadata.get("tier", "fact")),
                    created_at=created_at,
                    updated_at=(
                        datetime.fromisoformat(metadata["updated_at"])
                        if "updated_at" in metadata else now
                    ),
                ),
                similarity_score=score,
            ))

        ranked = score_and_rank(
            results=search_results,
            config=self.config.scoring,
            time_after=time_after,
            time_before=time_before,
            topic_weights=self.config.topics.weights or None,
            ymyl_config=self.config.ymyl if self.config.ymyl.enabled else None,
        )

        if self.config.enable_hierarchy and tier is None:
            preferred = classify_query(query)
            ranked = route_results(ranked, preferred)

        return ranked[:top_k]

    def summarize(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        force: bool = False,
    ) -> List[Memory]:
        return self._hierarchy.maybe_summarize(
            user_id=user_id,
            agent_id=agent_id,
            force=force,
        )

    def get(self, memory_id: str) -> Optional[Memory]:
        result = self.vector_store.get(memory_id)
        if result is None:
            return None
        _, metadata = result
        return Memory(
            id=memory_id,
            content=metadata.get("content", ""),
            user_id=metadata.get("user_id"),
            agent_id=metadata.get("agent_id"),
            importance=metadata.get("importance", 5.0),
            tier=MemoryTier(metadata.get("tier", "fact")),
        )

    def delete(self, memory_id: str) -> None:
        self.vector_store.delete(memory_id)

    def get_history(self, memory_id: str) -> List[HistoryEntry]:
        return self._history_store.get_history(memory_id)

    def count(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        tier: Optional[MemoryTier] = None,
    ) -> int:
        filters: Dict[str, str] = {}
        if user_id:
            filters["user_id"] = user_id
        if agent_id:
            filters["agent_id"] = agent_id
        if tier:
            filters["tier"] = tier.value
        items = self.vector_store.list_all(filters=filters or None, max_results=100000)
        return len(items)

    def export_json(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> str:
        filters: Dict[str, str] = {}
        if user_id:
            filters["user_id"] = user_id
        if agent_id:
            filters["agent_id"] = agent_id
        items = self.vector_store.list_all(filters=filters or None, max_results=100000)
        memories = []
        for id, metadata in items:
            memories.append({"id": id, **metadata})
        return json.dumps({"memories": memories, "count": len(memories)}, indent=2, default=str)

    def import_json(self, data: str) -> int:
        parsed = json.loads(data)
        imported = 0
        for item in parsed.get("memories", []):
            memory_id = item.get("id")
            content = item.get("content", "")
            if not content:
                continue
            existing = self.vector_store.get(memory_id) if memory_id else None
            if existing:
                continue
            embedding = self.embedder.embed(content)
            memory = Memory(
                id=memory_id or Memory().id,
                content=content,
                user_id=item.get("user_id"),
                agent_id=item.get("agent_id"),
                importance=item.get("importance", 5.0),
                tier=MemoryTier(item.get("tier", "fact")),
                content_hash=item.get("content_hash", ""),
            )
            metadata = {
                "content": memory.content,
                "user_id": memory.user_id,
                "agent_id": memory.agent_id,
                "run_id": memory.run_id,
                "tier": memory.tier.value,
                "importance": memory.importance,
                "content_hash": memory.content_hash,
                "created_at": item.get("created_at", memory.created_at.isoformat()),
                "updated_at": item.get("updated_at", memory.updated_at.isoformat()),
            }
            self.vector_store.insert(id=memory.id, vector=embedding, metadata=metadata)
            imported += 1
        return imported

    def _create_llm(self) -> BaseLLM:
        provider = self.config.llm.provider
        if provider == "openai":
            return OpenAILLM(self.config.llm)
        if provider == "anthropic":
            from widemem.providers.llm.anthropic import AnthropicLLM
            return AnthropicLLM(self.config.llm)
        if provider == "ollama":
            from widemem.providers.llm.ollama import OllamaLLM
            return OllamaLLM(self.config.llm)
        raise ValueError(
            f"Unknown LLM provider: {provider}. Supported: openai, anthropic, ollama"
        )

    def _create_embedder(self) -> BaseEmbedder:
        provider = self.config.embedding.provider
        if provider == "openai":
            return OpenAIEmbedder(self.config.embedding)
        if provider == "sentence-transformers":
            from widemem.providers.embeddings.sentence_transformers import (
                SentenceTransformerEmbedder,
            )
            return SentenceTransformerEmbedder(self.config.embedding)
        raise ValueError(
            f"Unknown embedding provider: {provider}. Supported: openai, sentence-transformers"
        )

    def _create_vector_store(self) -> BaseVectorStore:
        provider = self.config.vector_store.provider
        if provider == "faiss":
            return FAISSVectorStore(
                self.config.vector_store,
                dimensions=self.config.embedding.dimensions,
            )
        if provider == "qdrant":
            from widemem.storage.vector.qdrant_store import QdrantVectorStore
            return QdrantVectorStore(
                self.config.vector_store,
                dimensions=self.config.embedding.dimensions,
            )
        raise ValueError(
            f"Unknown vector store provider: {provider}. Supported: faiss, qdrant"
        )



from __future__ import annotations

from typing import List, Tuple

from widemem.core.types import Memory
from widemem.hierarchy.prompts import (
    GROUP_FACTS_PROMPT,
    GROUP_FACTS_SYSTEM,
    SUMMARIZE_FACTS_PROMPT,
    SUMMARIZE_FACTS_SYSTEM,
    SYNTHESIZE_THEME_PROMPT,
    SYNTHESIZE_THEME_SYSTEM,
)
from widemem.providers.llm.base import BaseLLM


class MemorySummarizer:
    def __init__(self, llm: BaseLLM) -> None:
        self.llm = llm

    def group_facts(self, facts: List[Memory]) -> List[Tuple[str, List[Memory]]]:
        if len(facts) < 2:
            return []

        facts_str = "\n".join(f"[{i}] {f.content}" for i, f in enumerate(facts))
        prompt = GROUP_FACTS_PROMPT.format(facts=facts_str)
        result = self.llm.generate_json(prompt, system=GROUP_FACTS_SYSTEM)

        groups = []
        for group in result.get("groups", []):
            label = group.get("label", "unknown")
            indices = group.get("fact_indices", [])
            members = [facts[i] for i in indices if i < len(facts)]
            if len(members) >= 2:
                groups.append((label, members))

        return groups

    def summarize_group(self, facts: List[Memory]) -> Tuple[str, float]:
        facts_str = "\n".join(f"- {f.content}" for f in facts)
        prompt = SUMMARIZE_FACTS_PROMPT.format(facts=facts_str)
        result = self.llm.generate_json(prompt, system=SUMMARIZE_FACTS_SYSTEM)

        summary = result.get("summary", "")
        importance = float(result.get("importance", 7.0))
        return summary, importance

    def synthesize_theme(self, summaries: List[Memory]) -> Tuple[str, float]:
        summaries_str = "\n".join(f"- {s.content}" for s in summaries)
        prompt = SYNTHESIZE_THEME_PROMPT.format(summaries=summaries_str)
        result = self.llm.generate_json(prompt, system=SYNTHESIZE_THEME_SYSTEM)

        theme = result.get("theme", "")
        importance = float(result.get("importance", 8.0))
        return theme, importance



from __future__ import annotations

import re
from typing import List, Optional, Tuple

from widemem.core.types import YMYLConfig

_YMYL_STRONG_PATTERNS = {
    "health": [
        r"\b(blood pressure|cholesterol level|diabetes diagnosis|mental health"
        r"|cancer treatment|anxiety disorder|depression diagnosis)\b",
    ],
    "medical": [
        r"\b(lab results|medical condition|chronic condition"
        r"|medical history|treatment plan|medical emergency)\b",
    ],
    "financial": [
        r"\b(bank account|savings account|credit score|credit card|mortgage rate"
        r"|loan payment|tax return|retirement fund|401k|IRA contribution"
        r"|pension plan|capital gains)\b",
    ],
    "legal": [
        r"\b(power of attorney|child custody|legal action|court order"
        r"|legal counsel|divorce proceedings|estate planning)\b",
    ],
    "safety": [
        r"\b(emergency contact|next of kin|blood type|epipen"
        r"|pacemaker|DNR order|medical alert)\b",
    ],
    "insurance": [
        r"\b(insurance policy|insurance premium|insurance claim"
        r"|health insurance|life insurance|coverage plan)\b",
    ],
    "tax": [
        r"\b(tax return|tax filing|W-2|1099|IRS audit"
        r"|capital gains tax|tax deduction)\b",
    ],
    "pharmaceutical": [
        r"\b(side effect|drug interaction|contraindication"
        r"|over-the-counter|prescription medication)\b",
    ],
}

_YMYL_WEAK_PATTERNS = {
    "health": [
        r"\b(doctor|hospital|diagnosis|symptom|allerg|medication|prescription|surgery|disease|illness|therapy|therapist|depression|anxiety|diabetes|cancer)\b",
    ],
    "medical": [
        r"\b(clinic|treatment|dosage|vaccine|chronic|acute|ambulance|nurse|specialist|referral|x-ray|MRI|scan)\b",
    ],
    "financial": [
        r"\b(bank|savings|invest|portfolio|stock|bond|mortgage|loan|debt|credit|salary|income|retirement|pension|budget|expense)\b",
    ],
    "legal": [
        r"\b(lawyer|attorney|lawsuit|court|contract|custody|divorce|estate|liability|compliance|regulation|verdict|settlement)\b",
    ],
    "safety": [
        r"\b(evacuation|flood|fire alarm)\b",
    ],
    "insurance": [
        r"\b(insurance|premium|deductible|coverage|claim|copay|beneficiary)\b",
    ],
    "tax": [
        r"\b(deduction|audit|write-off|exemption|filing)\b",
    ],
    "pharmaceutical": [
        r"\b(drug|pharmaceutical|dosage|pharmacist|prescription)\b",
    ],
}


class YMYLResult:
    __slots__ = ("category", "confidence")

    def __init__(self, category: Optional[str], confidence: str = "none") -> None:
        self.category = category
        self.confidence = confidence

    @property
    def is_ymyl(self) -> bool:
        return self.category is not None

    @property
    def is_strong(self) -> bool:
        return self.confidence == "strong"


def classify_ymyl_detailed(text: str, config: YMYLConfig) -> YMYLResult:
    if not config.enabled:
        return YMYLResult(None, "none")

    text_lower = text.lower()

    for category in config.categories:
        patterns = _YMYL_STRONG_PATTERNS.get(category, [])
        for pattern in patterns:
            if re.search(pattern, text_lower, re.IGNORECASE):
                return YMYLResult(category, "strong")

    weak_hits: List[Tuple[str, int]] = []
    for category in config.categories:
        patterns = _YMYL_WEAK_PATTERNS.get(category, [])
        count = 0
        for pattern in patterns:
            count += len(re.findall(pattern, text_lower, re.IGNORECASE))
        if count > 0:
            weak_hits.append((category, count))

    if not weak_hits:
        return YMYLResult(None, "none")

    weak_hits.sort(key=lambda x: x[1], reverse=True)
    best_category, best_count = weak_hits[0]

    if best_count >= 2:
        return YMYLResult(best_category, "strong")

    return YMYLResult(best_category, "weak")


def classify_ymyl(text: str, config: YMYLConfig) -> Optional[str]:
    result = classify_ymyl_detailed(text, config)
    return result.category


def classify_ymyl_batch(texts: List[str], config: YMYLConfig) -> List[Optional[str]]:
    return [classify_ymyl(t, config) for t in texts]


def is_ymyl(text: str, config: YMYLConfig) -> bool:
    return classify_ymyl(text, config) is not None


def is_ymyl_strong(text: str, config: YMYLConfig) -> bool:
    return classify_ymyl_detailed(text, config).is_strong



FACT_EXTRACTION_SYSTEM = """You are a fact extraction engine. Extract concrete, memorable facts from conversations.

Rules:
- Extract only factual statements, preferences, and personal details
- Each fact should be self-contained and understandable without context
- Preserve the original language of the conversation
- Rate each fact's importance from 1-10:
  - 1-3: Trivial mentions, passing comments
  - 4-6: Useful context, moderate preferences
  - 7-9: Core identity, strong preferences, critical information
  - 10: Essential, defining facts
{ymyl_instruction}
{custom_topics_instruction}
Respond with JSON only."""

FACT_EXTRACTION_PROMPT = """Extract facts from this conversation:

{text}

Respond with this exact JSON format:
{{"facts": [{{"content": "the fact", "importance": 7, "category": "general"}}, ...]}}

If there are no meaningful facts, respond with: {{"facts": []}}"""

YMYL_INSTRUCTION = (
    "\nCRITICAL: Facts about health, medical conditions, medications, finances, "
    "legal matters, insurance, taxes, or safety are EXTREMELY important. "
    "Rate these at 8-10 importance and tag their category accordingly. "
    'These are "Your Money or Your Life" (YMYL) facts that must never be lost.'
)

CUSTOM_TOPICS_INSTRUCTION = """
Pay special attention to facts about these topics: {topics}. Extract these with higher priority."""


def build_extraction_system(
    ymyl_enabled: bool = False,
    custom_topics: list = None,
) -> str:
    ymyl_inst = YMYL_INSTRUCTION if ymyl_enabled else ""
    topics_inst = ""
    if custom_topics:
        topics_inst = CUSTOM_TOPICS_INSTRUCTION.format(topics=", ".join(custom_topics))

    return FACT_EXTRACTION_SYSTEM.format(
        ymyl_instruction=ymyl_inst,
        custom_topics_instruction=topics_inst,
    )



from __future__ import annotations

from typing import Callable, List, Optional

from pydantic import BaseModel

from widemem.core.types import Fact, MemorySearchResult
from widemem.providers.llm.base import BaseLLM
from widemem.retrieval.prompts import (
    CONTRADICTION_DETECTION_PROMPT,
    CONTRADICTION_DETECTION_SYSTEM,
)
from widemem.utils.id_mapping import IDMapper


class Clarification(BaseModel):
    new_fact: str
    existing_content: str
    existing_memory_id: Optional[str] = None
    conflict_type: str  # "contradiction" or "ambiguity"
    question: str


ClarificationCallback = Callable[[List[Clarification]], Optional[List[str]]]


class ActiveRetrieval:
    def __init__(self, llm: BaseLLM, similarity_threshold: float = 0.6) -> None:
        self.llm = llm
        self.similarity_threshold = similarity_threshold

    def detect_conflicts(
        self,
        new_facts: List[Fact],
        existing_memories: List[MemorySearchResult],
    ) -> List[Clarification]:
        if not new_facts or not existing_memories:
            return []

        relevant = [
            m for m in existing_memories
            if m.similarity_score >= self.similarity_threshold
        ]
        if not relevant:
            return []

        mapper = IDMapper()
        for mem_result in relevant:
            mapper.add(mem_result.memory.id)

        existing_str = "\n".join(
            f"[{mapper.to_int(m.memory.id)}] {m.memory.content}"
            for m in relevant
        )

        all_clarifications: List[Clarification] = []

        for fact in new_facts:
            prompt = CONTRADICTION_DETECTION_PROMPT.format(
                new_fact=fact.content,
                existing_memories=existing_str,
            )

            try:
                result = self.llm.generate_json(prompt, system=CONTRADICTION_DETECTION_SYSTEM)
            except Exception:
                continue

            if not result.get("has_conflict", False):
                continue

            for conflict in result.get("conflicts", []):
                existing_id = conflict.get("existing_memory_id")
                real_id = mapper.to_uuid(int(existing_id)) if existing_id is not None else None

                all_clarifications.append(Clarification(
                    new_fact=fact.content,
                    existing_content=conflict.get("existing_content", ""),
                    existing_memory_id=real_id,
                    conflict_type=conflict.get("type", "ambiguity"),
                    question=conflict.get("question", ""),
                ))

        return all_clarifications



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

from widemem.core.types import VectorStoreConfig


class BaseVectorStore(ABC):
    def __init__(self, config: VectorStoreConfig) -> None:
        self.config = config

    @abstractmethod
    def insert(self, id: str, vector: list[float], metadata: dict[str, Any]) -> None:
        ...

    @abstractmethod
    def search(
        self,
        vector: list[float],
        top_k: int = 10,
        filters: dict[str, Any] | None = None,
    ) -> list[tuple[str, float, dict[str, Any]]]:
        """Returns list of (id, score, metadata) tuples."""
        ...

    @abstractmethod
    def update(self, id: str, vector: list[float], metadata: dict[str, Any]) -> None:
        ...

    @abstractmethod
    def delete(self, id: str) -> None:
        ...

    @abstractmethod
    def get(self, id: str) -> tuple[list[float], dict[str, Any]] | None:
        ...

    def list_all(
        self,
        filters: dict[str, Any] | None = None,
        max_results: int = 1000,
    ) -> list[tuple[str, dict[str, Any]]]:
        """Returns list of (id, metadata) for all matching entries. Override for efficiency."""
        raise NotImplementedError



from __future__ import annotations

from typing import List

from widemem.core.exceptions import ProviderError
from widemem.core.types import EmbeddingConfig
from widemem.providers.embeddings.base import BaseEmbedder


class SentenceTransformerEmbedder(BaseEmbedder):
    def __init__(self, config: EmbeddingConfig) -> None:
        super().__init__(config)
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError:
            raise ProviderError(
                "Install sentence-transformers: pip install widemem[sentence-transformers]"
            )
        self._model = SentenceTransformer(
            config.model or "all-MiniLM-L6-v2",
        )
        actual_dim = self._model.get_sentence_embedding_dimension()
        if config.dimensions and config.dimensions != actual_dim:
            self.config = EmbeddingConfig(
                provider=config.provider,
                model=config.model,
                api_key=config.api_key,
                dimensions=actual_dim,
            )

    def embed(self, text: str) -> List[float]:
        return self.embed_batch([text])[0]

    def embed_batch(self, texts: List[str]) -> List[List[float]]:
        if not texts:
            return []
        embeddings = self._model.encode(texts, normalize_embeddings=True)
        return [e.tolist() for e in embeddings]

    @property
    def dimensions(self) -> int:
        return self.config.dimensions



from __future__ import annotations

import sqlite3
import uuid
from datetime import datetime
from pathlib import Path

from widemem.core.types import HistoryEntry, MemoryAction


class HistoryStore:
    def __init__(self, db_path: str = "~/.widemem/history.db") -> None:
        path = Path(db_path).expanduser()
        path.parent.mkdir(parents=True, exist_ok=True)
        self.conn = sqlite3.connect(str(path))
        self._init_db()

    def _init_db(self) -> None:
        self.conn.execute("""
            CREATE TABLE IF NOT EXISTS history (
                id TEXT PRIMARY KEY,
                memory_id TEXT NOT NULL,
                action TEXT NOT NULL,
                old_content TEXT,
                new_content TEXT,
                timestamp TEXT NOT NULL
            )
        """)
        self.conn.execute("""
            CREATE INDEX IF NOT EXISTS idx_history_memory_id ON history(memory_id)
        """)
        self.conn.commit()

    def log(
        self,
        memory_id: str,
        action: MemoryAction,
        old_content: str | None = None,
        new_content: str | None = None,
    ) -> HistoryEntry:
        entry = HistoryEntry(
            id=str(uuid.uuid4()),
            memory_id=memory_id,
            action=action,
            old_content=old_content,
            new_content=new_content,
            timestamp=datetime.utcnow(),
        )
        self.conn.execute(
            "INSERT INTO history (id, memory_id, action, old_content, new_content, timestamp) VALUES (?, ?, ?, ?, ?, ?)",
            (entry.id, entry.memory_id, entry.action.value, entry.old_content, entry.new_content, entry.timestamp.isoformat()),
        )
        self.conn.commit()
        return entry

    def get_history(self, memory_id: str) -> list[HistoryEntry]:
        cursor = self.conn.execute(
            "SELECT id, memory_id, action, old_content, new_content, timestamp FROM history WHERE memory_id = ? ORDER BY timestamp",
            (memory_id,),
        )
        return [
            HistoryEntry(
                id=row[0],
                memory_id=row[1],
                action=MemoryAction(row[2]),
                old_content=row[3],
                new_content=row[4],
                timestamp=datetime.fromisoformat(row[5]),
            )
            for row in cursor.fetchall()
        ]

    def close(self) -> None:
        self.conn.close()



from __future__ import annotations

from openai import OpenAI

from widemem.core.exceptions import ProviderError
from widemem.core.types import EmbeddingConfig
from widemem.providers.embeddings.base import BaseEmbedder


class OpenAIEmbedder(BaseEmbedder):
    def __init__(self, config: EmbeddingConfig) -> None:
        super().__init__(config)
        self.client = OpenAI(api_key=config.api_key)

    def embed(self, text: str) -> list[float]:
        return self.embed_batch([text])[0]

    def embed_batch(self, texts: list[str]) -> list[list[float]]:
        if not texts:
            return []
        try:
            response = self.client.embeddings.create(
                model=self.config.model,
                input=texts,
                dimensions=self.config.dimensions,
            )
            return [item.embedding for item in response.data]
        except Exception as e:
            raise ProviderError(f"Embedding failed: {e}") from e



from __future__ import annotations

from datetime import datetime
from typing import Dict, Optional

from widemem.core.types import MemorySearchResult, ScoringConfig, YMYLConfig
from widemem.scoring.decay import compute_recency_score
from widemem.scoring.importance import normalize_importance
from widemem.scoring.topics import get_topic_boost
from widemem.scoring.ymyl import classify_ymyl_detailed


def score_and_rank(
    results: list[MemorySearchResult],
    config: ScoringConfig,
    now: Optional[datetime] = None,
    time_after: Optional[datetime] = None,
    time_before: Optional[datetime] = None,
    topic_weights: Optional[Dict[str, float]] = None,
    ymyl_config: Optional[YMYLConfig] = None,
) -> list[MemorySearchResult]:
    now = now or datetime.utcnow()
    ymyl_config = ymyl_config or YMYLConfig()

    scored = []
    for result in results:
        created_at = result.memory.created_at

        if time_after and created_at < time_after:
            continue
        if time_before and created_at > time_before:
            continue

        content = result.memory.content
        ymyl_result = classify_ymyl_detailed(content, ymyl_config) if ymyl_config.enabled else None
        mem_is_ymyl_strong = ymyl_result is not None and ymyl_result.is_strong

        if mem_is_ymyl_strong and ymyl_config.decay_immune:
            recency = 1.0
        else:
            recency = compute_recency_score(
                created_at=created_at,
                now=now,
                decay_function=config.decay_function,
                decay_rate=config.decay_rate,
            )

        importance = normalize_importance(result.memory.importance)

        final = (
            config.similarity_weight * result.similarity_score
            + config.importance_weight * importance
            + config.recency_weight * recency
        )

        if topic_weights:
            boost = get_topic_boost(content, topic_weights)
            final *= boost

        result.temporal_score = recency
        result.importance_score = importance
        result.final_score = final
        scored.append(result)

    scored.sort(key=lambda r: r.final_score, reverse=True)
    return scored



"""Tests for hierarchical memory: grouping, summarization, themes, and query routing."""

from __future__ import annotations

import tempfile
from typing import List

import numpy as np
import pytest

from widemem.core.types import (
    EmbeddingConfig,
    Fact,
    LLMConfig,
    Memory,
    MemoryConfig,
    MemorySearchResult,
    MemoryTier,
    VectorStoreConfig,
)
from widemem.extraction.base import BaseExtractor
from widemem.hierarchy.manager import HierarchyManager
from widemem.hierarchy.query_router import classify_query, route_results
from widemem.hierarchy.summarizer import MemorySummarizer
from widemem.providers.embeddings.base import BaseEmbedder
from widemem.providers.llm.base import BaseLLM
from widemem.storage.history import HistoryStore
from widemem.storage.vector.faiss_store import FAISSVectorStore


class MockLLM(BaseLLM):
    def __init__(self):
        super().__init__(LLMConfig())
        self.responses: list = []

    def set_responses(self, *responses):
        self.responses = list(responses)

    def generate(self, prompt, system=None):
        import json
        return json.dumps(self.generate_json(prompt, system))

    def generate_json(self, prompt, system=None):
        if self.responses:
            return self.responses.pop(0)
        return {}


class MockEmbedder(BaseEmbedder):
    def __init__(self, dimensions=64):
        super().__init__(EmbeddingConfig(dimensions=dimensions))

    def embed(self, text):
        rng = np.random.RandomState(hash(text) % 2**31)
        vec = rng.randn(self.config.dimensions).astype(np.float32)
        return (vec / np.linalg.norm(vec)).tolist()

    def embed_batch(self, texts):
        return [self.embed(t) for t in texts]


class DirectExtractor(BaseExtractor):
    def __init__(self):
        self.facts: List[Fact] = []

    def set_facts(self, *facts):
        self.facts = list(facts)

    def extract(self, text):
        if self.facts:
            return self.facts
        return [Fact(content=text, importance=5.0)]


@pytest.fixture
def tmp_dir():
    with tempfile.TemporaryDirectory() as d:
        yield d


@pytest.fixture
def mock_llm():
    return MockLLM()


@pytest.fixture
def embedder():
    return MockEmbedder(dimensions=64)


@pytest.fixture
def vector_store(tmp_dir):
    return FAISSVectorStore(VectorStoreConfig(path=f"{tmp_dir}/vectors"), dimensions=64)


@pytest.fixture
def history(tmp_dir):
    return HistoryStore(f"{tmp_dir}/history.db")


# --- Query Router Tests ---

class TestQueryRouter:
    def test_broad_query_routes_to_theme(self):
        assert classify_query("tell me about alice") == MemoryTier.THEME
        assert classify_query("who is alice") == MemoryTier.THEME
        assert classify_query("give me an overview of the project") == MemoryTier.THEME

    def test_specific_query_routes_to_fact(self):
        assert classify_query("where does alice live") == MemoryTier.FACT
        assert classify_query("when did she move") == MemoryTier.FACT
        assert classify_query("what is her job") == MemoryTier.FACT

    def test_short_query_routes_to_fact(self):
        assert classify_query("alice email") == MemoryTier.FACT
        assert classify_query("job title") == MemoryTier.FACT

    def test_medium_query_routes_to_summary(self):
        assert classify_query("alice work experience and skills in engineering") == MemoryTier.SUMMARY

    def test_route_results_preferred_tier(self):
        results = [
            MemorySearchResult(
                memory=Memory(content="fact1", tier=MemoryTier.FACT),
                final_score=0.9,
            ),
            MemorySearchResult(
                memory=Memory(content="summary1", tier=MemoryTier.SUMMARY),
                final_score=0.8,
            ),
            MemorySearchResult(
                memory=Memory(content="fact2", tier=MemoryTier.FACT),
                final_score=0.7,
            ),
            MemorySearchResult(
                memory=Memory(content="fact3", tier=MemoryTier.FACT),
                final_score=0.6,
            ),
        ]

        routed = route_results(results, MemoryTier.FACT)
        assert all(r.memory.tier == MemoryTier.FACT for r in routed)

    def test_route_results_fallback_when_insufficient(self):
        results = [
            MemorySearchResult(
                memory=Memory(content="theme1", tier=MemoryTier.THEME),
                final_score=0.9,
            ),
            MemorySearchResult(
                memory=Memory(content="fact1", tier=MemoryTier.FACT),
                final_score=0.8,
            ),
        ]

        routed = route_results(results, MemoryTier.THEME, min_results=3)
        assert len(routed) == 2  # only 2 total available
        assert routed[0].memory.tier == MemoryTier.THEME


# --- Summarizer Tests ---

class TestSummarizer:
    def test_group_facts(self, mock_llm):
        summarizer = MemorySummarizer(mock_llm)
        facts = [
            Memory(content="Lives in Berlin"),
            Memory(content="Works at Google"),
            Memory(content="Enjoys hiking"),
            Memory(content="Speaks German"),
        ]

        mock_llm.set_responses({
            "groups": [
                {"label": "Location & Language", "fact_indices": [0, 3]},
                {"label": "Work & Hobbies", "fact_indices": [1, 2]},
            ]
        })

        groups = summarizer.group_facts(facts)
        assert len(groups) == 2
        assert groups[0][0] == "Location & Language"
        assert len(groups[0][1]) == 2
        assert groups[1][0] == "Work & Hobbies"

    def test_group_facts_skips_small_groups(self, mock_llm):
        summarizer = MemorySummarizer(mock_llm)
        facts = [Memory(content="fact1"), Memory(content="fact2"), Memory(content="fact3")]

        mock_llm.set_responses({
            "groups": [
                {"label": "big group", "fact_indices": [0, 1]},
                {"label": "solo", "fact_indices": [2]},  # only 1 fact
            ]
        })

        groups = summarizer.group_facts(facts)
        assert len(groups) == 1

    def test_summarize_group(self, mock_llm):
        summarizer = MemorySummarizer(mock_llm)
        facts = [Memory(content="Lives in Berlin"), Memory(content="Speaks German")]

        mock_llm.set_responses({"summary": "Lives in Berlin and speaks German.", "importance": 8.0})

        text, importance = summarizer.summarize_group(facts)
        assert text == "Lives in Berlin and speaks German."
        assert importance == 8.0

    def test_synthesize_theme(self, mock_llm):
        summarizer = MemorySummarizer(mock_llm)
        summaries = [
            Memory(content="Lives in Berlin and speaks German."),
            Memory(content="Works at Google as a senior engineer."),
        ]

        mock_llm.set_responses({
            "theme": "A German-speaking senior engineer at Google based in Berlin.",
            "importance": 9.0,
        })

        text, importance = summarizer.synthesize_theme(summaries)
        assert "Berlin" in text
        assert importance == 9.0

    def test_group_facts_too_few(self, mock_llm):
        summarizer = MemorySummarizer(mock_llm)
        groups = summarizer.group_facts([Memory(content="only one")])
        assert groups == []


# --- Hierarchy Manager Tests ---

class TestHierarchyManager:
    def _insert_facts(self, vector_store, embedder, history, facts, user_id="alice"):
        from widemem.utils.hashing import content_hash
        memories = []
        for text in facts:
            mem = Memory(
                content=text,
                user_id=user_id,
                tier=MemoryTier.FACT,
                importance=7.0,
                content_hash=content_hash(text),
            )
            embedding = embedder.embed(text)
            vector_store.insert(
                id=mem.id,
                vector=embedding,
                metadata={
                    "content": mem.content,
                    "user_id": mem.user_id,
                    "tier": mem.tier.value,
                    "importance": mem.importance,
                    "content_hash": mem.content_hash,
                    "created_at": mem.created_at.isoformat(),
                    "updated_at": mem.updated_at.isoformat(),
                },
            )
            memories.append(mem)
        return memories

    def test_below_threshold_no_summarize(self, mock_llm, embedder, vector_store, history):
        manager = HierarchyManager(
            summarizer=MemorySummarizer(mock_llm),
            embedder=embedder,
            vector_store=vector_store,
            history=history,
            summarize_threshold=10,
        )

        self._insert_facts(vector_store, embedder, history, [
            "Likes coffee", "Lives in Berlin",
        ])

        result = manager.maybe_summarize(user_id="alice")
        assert result == []

    def test_force_summarize(self, mock_llm, embedder, vector_store, history):
        manager = HierarchyManager(
            summarizer=MemorySummarizer(mock_llm),
            embedder=embedder,
            vector_store=vector_store,
            history=history,
            summarize_threshold=100,
            theme_threshold=100,
        )

        self._insert_facts(vector_store, embedder, history, [
            "Lives in Berlin", "Speaks German",
            "Works at Google", "Senior engineer",
        ])

        mock_llm.set_responses(
            # group_facts
            {"groups": [{"label": "Location", "fact_indices": [0, 1]}, {"label": "Work", "fact_indices": [2, 3]}]},
            # summarize_group for Location
            {"summary": "Based in Berlin, speaks German.", "importance": 8.0},
            # summarize_group for Work
            {"summary": "Senior engineer at Google.", "importance": 8.0},
        )

        result = manager.maybe_summarize(user_id="alice", force=True)
        assert len(result) == 2
        assert all(m.tier == MemoryTier.SUMMARY for m in result)

    def test_theme_generated_when_enough_summaries(self, mock_llm, embedder, vector_store, history):
        manager = HierarchyManager(
            summarizer=MemorySummarizer(mock_llm),
            embedder=embedder,
            vector_store=vector_store,
            history=history,
            summarize_threshold=1,
            theme_threshold=3,
        )

        self._insert_facts(vector_store, embedder, history, [
            "Lives in Berlin", "Speaks German",
            "Works at Google", "Senior engineer",
            "Enjoys hiking", "Runs marathons",
        ])

        mock_llm.set_responses(
            # group_facts
            {"groups": [
                {"label": "Location", "fact_indices": [0, 1]},
                {"label": "Work", "fact_indices": [2, 3]},
                {"label": "Sports", "fact_indices": [4, 5]},
            ]},
            # 3 summarize_group calls
            {"summary": "Based in Berlin, German speaker.", "importance": 8.0},
            {"summary": "Senior engineer at Google.", "importance": 8.0},
            {"summary": "Active outdoors person, hiker and runner.", "importance": 6.0},
            # synthesize_theme
            {"theme": "A German-speaking senior Google engineer in Berlin who enjoys outdoor sports.", "importance": 9.0},
        )

        result = manager.maybe_summarize(user_id="alice", force=True)
        summaries = [m for m in result if m.tier == MemoryTier.SUMMARY]
        themes = [m for m in result if m.tier == MemoryTier.THEME]
        assert len(summaries) == 3
        assert len(themes) == 1
        assert "Berlin" in themes[0].content


# --- Integration Test ---

class TestHierarchyIntegration:
    def test_search_with_hierarchy_routing(self, tmp_dir):
        from widemem.core.memory import WideMemory

        mock_llm = MockLLM()
        emb = MockEmbedder(dimensions=64)
        vs = FAISSVectorStore(VectorStoreConfig(path=f"{tmp_dir}/vectors"), dimensions=64)

        config = MemoryConfig(
            history_db_path=f"{tmp_dir}/history.db",
            embedding=EmbeddingConfig(dimensions=64),
            enable_hierarchy=True,
        )

        mem = WideMemory(config=config, llm=mock_llm, embedder=emb, vector_store=vs)
        extractor = DirectExtractor()
        mem.pipeline.extractor = extractor

        # Add facts
        for fact in ["Lives in Berlin", "Works at Google", "Speaks German", "Enjoys hiking"]:
            extractor.set_facts(Fact(content=fact, importance=7.0))
            mem.add(fact, user_id="alice")

        # Summarize
        mock_llm.set_responses(
            {"groups": [{"label": "Profile", "fact_indices": [0, 1, 2, 3]}]},
            {"summary": "German-speaking Googler in Berlin who hikes.", "importance": 8.0},
        )
        summaries = mem.summarize(user_id="alice", force=True)
        assert len(summaries) >= 1

        # Search with specific query - should prefer facts
        results = mem.search("where does alice live", user_id="alice")
        assert len(results) >= 1

        # Search with broad query - should prefer themes/summaries if available
        results = mem.search("tell me about alice", user_id="alice")
        assert len(results) >= 1

    def test_search_tier_filter(self, tmp_dir):
        from widemem.core.memory import WideMemory

        mock_llm = MockLLM()
        emb = MockEmbedder(dimensions=64)
        vs = FAISSVectorStore(VectorStoreConfig(path=f"{tmp_dir}/vectors"), dimensions=64)

        config = MemoryConfig(
            history_db_path=f"{tmp_dir}/history.db",
            embedding=EmbeddingConfig(dimensions=64),
        )

        mem = WideMemory(config=config, llm=mock_llm, embedder=emb, vector_store=vs)
        extractor = DirectExtractor()
        mem.pipeline.extractor = extractor

        extractor.set_facts(Fact(content="Lives in Berlin", importance=7.0))
        mem.add("Lives in Berlin", user_id="alice")

        # Search for facts only
        results = mem.search("Berlin", user_id="alice", tier=MemoryTier.FACT)
        assert len(results) >= 1
        assert all(r.memory.tier == MemoryTier.FACT for r in results)

        # Search for summaries only - should be empty
        results = mem.search("Berlin", user_id="alice", tier=MemoryTier.SUMMARY)
        assert len(results) == 0



#!/usr/bin/env python3
"""Train a self-supervised fact extraction model from collected LLM extractions.

Usage:
    python scripts/train_extractor.py --data extractions.jsonl --output models/extractor

Requires: pip install transformers datasets torch
"""
from __future__ import annotations

import argparse
import json
from pathlib import Path


def load_data(path: str):
    samples = []
    with open(path) as f:
        for line in f:
            item = json.loads(line)
            input_text = item["input"]
            output = json.dumps({"facts": item["output"]})
            samples.append({"input": input_text, "output": output})
    return samples


def train(data_path: str, output_dir: str, base_model: str, epochs: int, batch_size: int):
    try:
        from datasets import Dataset
        from transformers import (
            AutoModelForSeq2SeqLM,
            AutoTokenizer,
            Seq2SeqTrainer,
            Seq2SeqTrainingArguments,
        )
    except ImportError:
        print("Error: Install training dependencies:")
        print("  pip install transformers datasets torch")
        return

    samples = load_data(data_path)
    print(f"Loaded {len(samples)} training samples")

    if len(samples) < 50:
        print(f"Warning: Only {len(samples)} samples. Recommend at least 500 for good quality.")

    tokenizer = AutoTokenizer.from_pretrained(base_model)
    model = AutoModelForSeq2SeqLM.from_pretrained(base_model)

    def preprocess(examples):
        inputs = tokenizer(
            examples["input"],
            max_length=512,
            truncation=True,
            padding="max_length",
        )
        targets = tokenizer(
            examples["output"],
            max_length=256,
            truncation=True,
            padding="max_length",
        )
        inputs["labels"] = targets["input_ids"]
        return inputs

    dataset = Dataset.from_list(samples)
    split = dataset.train_test_split(test_size=0.1, seed=42)
    train_ds = split["train"].map(preprocess, batched=True, remove_columns=["input", "output"])
    eval_ds = split["test"].map(preprocess, batched=True, remove_columns=["input", "output"])

    training_args = Seq2SeqTrainingArguments(
        output_dir=output_dir,
        num_train_epochs=epochs,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        eval_strategy="epoch",
        save_strategy="epoch",
        logging_steps=10,
        load_best_model_at_end=True,
        predict_with_generate=True,
        fp16=False,
    )

    trainer = Seq2SeqTrainer(
        model=model,
        args=training_args,
        train_dataset=train_ds,
        eval_dataset=eval_ds,
        tokenizer=tokenizer,
    )

    trainer.train()
    trainer.save_model(output_dir)
    tokenizer.save_pretrained(output_dir)
    print(f"Model saved to {output_dir}")


def export_data(db_path: str, output_path: str):
    from widemem.extraction.collector import ExtractionCollector
    collector = ExtractionCollector(db_path)
    count = collector.export(output_path)
    collector.close()
    print(f"Exported {count} samples to {output_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Train widemem fact extraction model")
    sub = parser.add_subparsers(dest="command")

    export_parser = sub.add_parser("export", help="Export training data from collector DB")
    export_parser.add_argument("--db", default="~/.widemem/extractions.db")
    export_parser.add_argument("--output", default="extractions.jsonl")

    train_parser = sub.add_parser("train", help="Train the extraction model")
    train_parser.add_argument("--data", required=True, help="Path to JSONL training data")
    train_parser.add_argument("--output", default="models/extractor", help="Output model directory")
    train_parser.add_argument("--base-model", default="google/flan-t5-small", help="Base model to fine-tune")
    train_parser.add_argument("--epochs", type=int, default=5)
    train_parser.add_argument("--batch-size", type=int, default=8)

    args = parser.parse_args()

    if args.command == "export":
        export_data(args.db, args.output)
    elif args.command == "train":
        train(args.data, args.output, args.base_model, args.epochs, args.batch_size)
    else:
        parser.print_help()



from __future__ import annotations

from typing import List

from widemem.core.types import MemorySearchResult, MemoryTier

BROAD_KEYWORDS = {
    "tell me about", "describe", "who is", "what do you know about",
    "summary", "overview", "profile", "background", "everything about",
}

SPECIFIC_KEYWORDS = {
    "where", "when", "what is", "how old", "how much", "how many",
    "does", "did", "is", "was", "has", "have", "which", "name",
}


def classify_query(query: str) -> MemoryTier:
    q = query.lower().strip()

    for kw in BROAD_KEYWORDS:
        if kw in q:
            return MemoryTier.THEME

    for kw in SPECIFIC_KEYWORDS:
        if q.startswith(kw):
            return MemoryTier.FACT

    word_count = len(q.split())
    if word_count <= 4:
        return MemoryTier.FACT

    return MemoryTier.SUMMARY


def route_results(
    results: List[MemorySearchResult],
    preferred_tier: MemoryTier,
    min_results: int = 3,
) -> List[MemorySearchResult]:
    tier_results = [r for r in results if r.memory.tier == preferred_tier]

    if len(tier_results) >= min_results:
        return tier_results

    if preferred_tier == MemoryTier.THEME:
        fallback_order = [MemoryTier.SUMMARY, MemoryTier.FACT]
    elif preferred_tier == MemoryTier.SUMMARY:
        fallback_order = [MemoryTier.THEME, MemoryTier.FACT]
    else:
        fallback_order = [MemoryTier.SUMMARY, MemoryTier.THEME]

    combined = list(tier_results)
    for tier in fallback_order:
        if len(combined) >= min_results:
            break
        combined.extend(r for r in results if r.memory.tier == tier)

    return combined



from widemem.core.memory import WideMemory
from widemem.core.pipeline import AddResult
from widemem.core.types import MemoryConfig
from widemem.retrieval.active import Clarification

__all__ = ["WideMemory", "MemoryConfig", "AddResult", "Clarification"]
__version__ = "1.3.0"



from __future__ import annotations

from typing import Dict, Optional


def get_topic_boost(
    content: str,
    topic_weights: Dict[str, float],
) -> float:
    if not topic_weights:
        return 1.0

    content_lower = content.lower()
    best_boost = 1.0

    for topic, weight in topic_weights.items():
        if topic.lower() in content_lower:
            best_boost = max(best_boost, weight)

    return best_boost


def get_topic_label(
    content: str,
    topic_weights: Dict[str, float],
) -> Optional[str]:
    if not topic_weights:
        return None

    content_lower = content.lower()
    for topic in topic_weights:
        if topic.lower() in content_lower:
            return topic

    return None



from __future__ import annotations

import time
from abc import ABC

from widemem.core.exceptions import ProviderError
from widemem.core.types import LLMConfig


class BaseLLM(ABC):
    def __init__(self, config: LLMConfig, max_retries: int = 3, retry_delay: float = 1.0) -> None:
        self.config = config
        self.max_retries = max_retries
        self.retry_delay = retry_delay

    def _generate(self, prompt: str, system: str | None = None) -> str:
        raise NotImplementedError

    def _generate_json(self, prompt: str, system: str | None = None) -> dict:
        raise NotImplementedError

    def generate(self, prompt: str, system: str | None = None) -> str:
        return self._retry(self._generate, prompt, system)

    def generate_json(self, prompt: str, system: str | None = None) -> dict:
        return self._retry(self._generate_json, prompt, system)

    def _retry(self, fn, prompt: str, system: str | None):
        last_error = None
        for attempt in range(self.max_retries):
            try:
                return fn(prompt, system)
            except ProviderError:
                raise
            except Exception as e:
                last_error = e
                if attempt < self.max_retries - 1:
                    time.sleep(self.retry_delay * (2 ** attempt))
        raise ProviderError(
            f"Failed after {self.max_retries} retries: {last_error}"
        ) from last_error



from __future__ import annotations

import json
from typing import List, Optional

from widemem.core.types import Fact
from widemem.extraction.base import BaseExtractor


class SelfSupervisedExtractor(BaseExtractor):
    """Extraction using a fine-tuned small model with LLM fallback.

    The model is expected to take input text and return JSON:
    {"facts": [{"content": "...", "importance": 7.0, "confidence": 0.95}, ...]}

    When confidence is below threshold, falls back to the LLM extractor.
    """

    def __init__(
        self,
        model_path: Optional[str] = None,
        fallback: Optional[BaseExtractor] = None,
        confidence_threshold: float = 0.7,
    ) -> None:
        self.model_path = model_path
        self.fallback = fallback
        self.confidence_threshold = confidence_threshold
        self._model = None

        if model_path:
            self._load_model(model_path)

    def _load_model(self, path: str) -> None:
        try:
            from transformers import pipeline as hf_pipeline
            self._model = hf_pipeline(
                "text2text-generation",
                model=path,
                max_new_tokens=512,
            )
        except ImportError:
            pass
        except Exception:
            pass

    def extract(self, text: str) -> List[Fact]:
        if self._model is None:
            if self.fallback:
                return self.fallback.extract(text)
            return []

        try:
            result = self._model(text)[0]["generated_text"]
            parsed = json.loads(result)
        except (json.JSONDecodeError, KeyError, IndexError):
            if self.fallback:
                return self.fallback.extract(text)
            return []

        facts = []
        low_confidence = False

        for item in parsed.get("facts", []):
            confidence = float(item.get("confidence", 0.5))
            if confidence < self.confidence_threshold:
                low_confidence = True
                break
            facts.append(Fact(
                content=item["content"],
                importance=float(item.get("importance", 5.0)),
            ))

        if low_confidence and self.fallback:
            return self.fallback.extract(text)

        return facts

    @property
    def is_loaded(self) -> bool:
        return self._model is not None



from __future__ import annotations

import json

from widemem.core.exceptions import ProviderError
from widemem.core.types import LLMConfig
from widemem.providers.llm.base import BaseLLM


class AnthropicLLM(BaseLLM):
    def __init__(self, config: LLMConfig) -> None:
        super().__init__(config)
        try:
            from anthropic import Anthropic
        except ImportError:
            raise ProviderError("Install anthropic: pip install widemem[anthropic]")
        self.client = Anthropic(api_key=config.api_key)

    def _generate(self, prompt: str, system: str | None = None) -> str:
        kwargs = {
            "model": self.config.model,
            "max_tokens": self.config.max_tokens,
            "messages": [{"role": "user", "content": prompt}],
        }
        if system:
            kwargs["system"] = system
        if self.config.temperature > 0:
            kwargs["temperature"] = self.config.temperature

        response = self.client.messages.create(**kwargs)
        if not response.content:
            raise ProviderError("Anthropic returned empty response")
        return response.content[0].text

    def _generate_json(self, prompt: str, system: str | None = None) -> dict:
        json_prompt = prompt + "\n\nRespond with valid JSON only."
        text = self._generate(json_prompt, system=system)

        text = text.strip()
        if text.startswith("```json"):
            text = text[7:]
        if text.startswith("```"):
            text = text[3:]
        if text.endswith("```"):
            text = text[:-3]
        text = text.strip()

        try:
            return json.loads(text)
        except json.JSONDecodeError as e:
            raise ProviderError(f"Anthropic returned invalid JSON: {e}") from e



BATCH_CONFLICT_RESOLUTION_SYSTEM = (
    "You are a memory conflict resolution engine. "
    "Given new facts and existing memories, decide what actions to take.\n\n"
    "Actions:\n"
    "- ADD: The fact is new information not captured by any existing memory\n"
    "- UPDATE: The fact contradicts, refines, or supersedes an existing memory "
    "(provide the target memory ID)\n"
    "- DELETE: An existing memory is now invalid based on the new facts "
    "(provide the target memory ID)\n"
    "- NONE: The fact is already captured by an existing memory, no action needed\n\n"
    "Rules:\n"
    "- Be conservative: prefer NONE over ADD if a fact is already well-captured\n"
    "- Detect contradictions: \"moved to Berlin\" should UPDATE \"lives in Paris\"\n"
    "- Detect refinements: \"works as a senior engineer at Google\" should UPDATE "
    "\"works at Google\"\n"
    "- One action per new fact\n"
    "- Respond with JSON only."
)

BATCH_CONFLICT_RESOLUTION_PROMPT = (
    "New facts to process:\n{new_facts}\n\n"
    "Existing memories:\n{existing_memories}\n\n"
    "For each new fact, decide the action. Respond with this exact JSON format:\n"
    '{{\"actions\": [{{\"fact_index\": 0, \"action\": \"add\", '
    '\"target_id\": null, \"importance\": 7}}, '
    '{{\"fact_index\": 1, \"action\": \"update\", '
    '\"target_id\": 3, \"importance\": 8}}, ...]}}\n\n'
    "fact_index corresponds to the index of the new fact (0-based).\n"
    "target_id is the ID of the existing memory to update/delete (null for add/none).\n"
    "importance is the importance score (1-10) for the resulting memory."
)



class WideMemError(Exception):
    pass


class ProviderError(WideMemError):
    pass


class ExtractionError(WideMemError):
    pass


class ConflictResolutionError(WideMemError):
    pass


class StorageError(WideMemError):
    pass



"""Tests for YMYL classification, topic boosting, and integration."""

from __future__ import annotations

from datetime import datetime, timedelta

import pytest

from widemem.core.types import (
    Memory,
    MemorySearchResult,
    ScoringConfig,
    YMYLConfig,
)
from widemem.retrieval.temporal import score_and_rank
from widemem.scoring.topics import get_topic_boost, get_topic_label
from widemem.scoring.ymyl import (
    classify_ymyl,
    classify_ymyl_batch,
    classify_ymyl_detailed,
    is_ymyl,
    is_ymyl_strong,
)


class TestYMYLClassifier:
    def setup_method(self):
        self.config = YMYLConfig(enabled=True)

    def test_strong_health(self):
        result = classify_ymyl_detailed("diabetes diagnosis confirmed", self.config)
        assert result.category == "health"
        assert result.is_strong

    def test_strong_financial(self):
        result = classify_ymyl_detailed("opened a savings account at the bank", self.config)
        assert result.category == "financial"
        assert result.is_strong

    def test_strong_safety(self):
        result = classify_ymyl_detailed("my blood type is O+", self.config)
        assert result.category == "safety"
        assert result.is_strong

    def test_strong_legal(self):
        result = classify_ymyl_detailed("filed for child custody", self.config)
        assert result.category == "legal"
        assert result.is_strong

    def test_strong_insurance(self):
        result = classify_ymyl_detailed("my insurance premium went up", self.config)
        assert result.category == "insurance"
        assert result.is_strong

    def test_strong_tax(self):
        result = classify_ymyl_detailed("filing my W-2 this week", self.config)
        assert result.category == "tax"
        assert result.is_strong

    def test_strong_pharmaceutical(self):
        result = classify_ymyl_detailed("checking for drug interaction risks", self.config)
        assert result.category == "pharmaceutical"
        assert result.is_strong

    def test_strong_via_two_weak_hits(self):
        result = classify_ymyl_detailed("doctor prescribed medication", self.config)
        assert result.category == "health"
        assert result.is_strong

    def test_weak_single_keyword(self):
        result = classify_ymyl_detailed("went to the doctor", self.config)
        assert result.category == "health"
        assert result.confidence == "weak"
        assert not result.is_strong

    def test_weak_bank_alone(self):
        result = classify_ymyl_detailed("walked by the bank", self.config)
        assert result.category == "financial"
        assert result.confidence == "weak"

    def test_false_positive_prevention(self):
        result = classify_ymyl_detailed("I like pizza", self.config)
        assert result.category is None
        assert result.confidence == "none"

        result = classify_ymyl_detailed("going to the park", self.config)
        assert result.category is None

    def test_is_ymyl_strong_helper(self):
        assert is_ymyl_strong("diabetes diagnosis confirmed", self.config) is True
        assert is_ymyl_strong("walked by the bank", self.config) is False
        assert is_ymyl_strong("pizza for dinner", self.config) is False

    def test_is_ymyl_any(self):
        assert is_ymyl("walked by the doctor", self.config) is True
        assert is_ymyl("pizza for dinner", self.config) is False

    def test_disabled(self):
        config = YMYLConfig(enabled=False)
        assert classify_ymyl("diabetes diagnosis", config) is None

    def test_limited_categories(self):
        config = YMYLConfig(enabled=True, categories=["health"])
        assert classify_ymyl("doctor appointment", config) == "health"
        assert classify_ymyl("mortgage rate is 6%", config) is None

    def test_batch_classify(self):
        texts = ["doctor prescribed medication", "pizza", "savings account at bank"]
        results = classify_ymyl_batch(texts, self.config)
        assert results[0] == "health"
        assert results[1] is None
        assert results[2] == "financial"

    def test_classify_ymyl_returns_category(self):
        assert classify_ymyl("401k contribution plan", self.config) == "financial"
        assert classify_ymyl("emergency contact is Jane", self.config) == "safety"


class TestTopicBoost:
    def test_no_weights(self):
        assert get_topic_boost("anything", {}) == 1.0

    def test_matching_topic(self):
        weights = {"python": 2.0, "rust": 1.5}
        assert get_topic_boost("I love Python programming", weights) == 2.0

    def test_no_match(self):
        weights = {"python": 2.0}
        assert get_topic_boost("I love JavaScript", weights) == 1.0

    def test_case_insensitive(self):
        weights = {"Python": 2.0}
        assert get_topic_boost("python is great", weights) == 2.0

    def test_best_boost_wins(self):
        weights = {"python": 2.0, "programming": 3.0}
        assert get_topic_boost("python programming", weights) == 3.0

    def test_topic_label(self):
        weights = {"python": 2.0, "rust": 1.5}
        assert get_topic_label("learning python", weights) == "python"
        assert get_topic_label("learning java", weights) is None

    def test_topic_label_empty(self):
        assert get_topic_label("anything", {}) is None


class TestYMYLDecayImmunity:
    def test_strong_ymyl_memory_no_decay(self):
        config = ScoringConfig()
        ymyl_config = YMYLConfig(enabled=True, decay_immune=True)
        now = datetime(2026, 3, 8)

        results = [
            MemorySearchResult(
                memory=Memory(
                    content="has diabetes diagnosis confirmed by doctor",
                    importance=9.0,
                    created_at=now - timedelta(days=365),
                ),
                similarity_score=0.8,
            ),
            MemorySearchResult(
                memory=Memory(
                    content="likes pizza",
                    importance=3.0,
                    created_at=now - timedelta(days=365),
                ),
                similarity_score=0.8,
            ),
        ]

        ranked = score_and_rank(results, config, now=now, ymyl_config=ymyl_config)
        ymyl_result = next(r for r in ranked if "diabetes" in r.memory.content)
        normal_result = next(r for r in ranked if "pizza" in r.memory.content)

        assert ymyl_result.temporal_score == 1.0
        assert normal_result.temporal_score < 1.0
        assert ymyl_result.final_score > normal_result.final_score

    def test_weak_ymyl_still_decays(self):
        config = ScoringConfig()
        ymyl_config = YMYLConfig(enabled=True, decay_immune=True)
        now = datetime(2026, 3, 8)

        results = [
            MemorySearchResult(
                memory=Memory(
                    content="walked by the bank",
                    importance=3.0,
                    created_at=now - timedelta(days=365),
                ),
                similarity_score=0.8,
            ),
        ]

        ranked = score_and_rank(results, config, now=now, ymyl_config=ymyl_config)
        assert ranked[0].temporal_score < 1.0

    def test_ymyl_decay_immune_false(self):
        config = ScoringConfig()
        ymyl_config = YMYLConfig(enabled=True, decay_immune=False)
        now = datetime(2026, 3, 8)

        results = [
            MemorySearchResult(
                memory=Memory(
                    content="has diabetes diagnosis from the doctor",
                    importance=9.0,
                    created_at=now - timedelta(days=365),
                ),
                similarity_score=0.8,
            ),
        ]

        ranked = score_and_rank(results, config, now=now, ymyl_config=ymyl_config)
        assert ranked[0].temporal_score < 1.0


class TestTopicBoostInScoring:
    def test_topic_boost_applied(self):
        config = ScoringConfig()
        now = datetime(2026, 3, 8)
        topic_weights = {"python": 2.0}

        results = [
            MemorySearchResult(
                memory=Memory(
                    content="loves python programming",
                    importance=5.0,
                    created_at=now - timedelta(days=1),
                ),
                similarity_score=0.8,
            ),
            MemorySearchResult(
                memory=Memory(
                    content="likes hiking on weekends",
                    importance=5.0,
                    created_at=now - timedelta(days=1),
                ),
                similarity_score=0.8,
            ),
        ]

        ranked = score_and_rank(results, config, now=now, topic_weights=topic_weights)
        python_result = next(r for r in ranked if "python" in r.memory.content)
        hiking_result = next(r for r in ranked if "hiking" in r.memory.content)

        assert python_result.final_score > hiking_result.final_score
        assert python_result.final_score == pytest.approx(hiking_result.final_score * 2.0, rel=0.01)

    def test_no_topic_weights(self):
        config = ScoringConfig()
        now = datetime(2026, 3, 8)

        results = [
            MemorySearchResult(
                memory=Memory(
                    content="loves python",
                    importance=5.0,
                    created_at=now,
                ),
                similarity_score=0.8,
            ),
        ]

        ranked = score_and_rank(results, config, now=now)
        ranked_with_topics = score_and_rank(results, config, now=now, topic_weights=None)
        assert ranked[0].final_score == ranked_with_topics[0].final_score


class TestEdgeCases:
    def test_dimension_validation(self):
        from widemem.core.types import VectorStoreConfig
        from widemem.storage.vector.faiss_store import FAISSVectorStore

        store = FAISSVectorStore(VectorStoreConfig(), dimensions=4)
        with pytest.raises(ValueError, match="dimension mismatch"):
            store.insert("id1", [0.1, 0.2], {"content": "too short"})

    def test_dimension_validation_on_search(self):
        from widemem.core.types import VectorStoreConfig
        from widemem.storage.vector.faiss_store import FAISSVectorStore

        store = FAISSVectorStore(VectorStoreConfig(), dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "ok"})
        with pytest.raises(ValueError, match="dimension mismatch"):
            store.search([0.1, 0.2], top_k=1)

    def test_conflict_resolver_fallback_on_bad_json(self):
        from widemem.conflict.batch_resolver import BatchConflictResolver
        from widemem.core.types import Fact, LLMConfig
        from widemem.providers.llm.base import BaseLLM

        class CrashingLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig())
            def generate(self, prompt, system=None):
                raise RuntimeError("LLM exploded")
            def generate_json(self, prompt, system=None):
                raise RuntimeError("LLM exploded")

        resolver = BatchConflictResolver(CrashingLLM())
        facts = [Fact(content="test fact", importance=5.0)]
        existing = [MemorySearchResult(
            memory=Memory(content="old fact"),
            similarity_score=0.9,
        )]
        actions = resolver.resolve(facts, existing)
        assert len(actions) == 1
        assert actions[0].action.value == "add"
        assert actions[0].fact == "test fact"

    def test_context_manager(self):
        import tempfile

        from widemem.core.memory import WideMemory
        from widemem.core.types import EmbeddingConfig, LLMConfig, MemoryConfig, VectorStoreConfig

        class DummyLLM:
            config = LLMConfig()
            def generate(self, p, system=None): return "{}"
            def generate_json(self, p, system=None): return {"facts": []}

        class DummyEmbedder:
            config = EmbeddingConfig(dimensions=4)
            dimensions = 4
            def embed(self, t): return [0.1, 0.2, 0.3, 0.4]
            def embed_batch(self, ts): return [[0.1, 0.2, 0.3, 0.4]] * len(ts)

        with tempfile.TemporaryDirectory() as d:
            config = MemoryConfig(history_db_path=f"{d}/h.db")
            from widemem.storage.vector.faiss_store import FAISSVectorStore
            vs = FAISSVectorStore(VectorStoreConfig(), dimensions=4)
            with WideMemory(config=config, llm=DummyLLM(), embedder=DummyEmbedder(), vector_store=vs):
                pass

    def test_duplicate_content_skipped(self):
        from widemem.core.types import VectorStoreConfig
        from widemem.storage.vector.faiss_store import FAISSVectorStore
        from widemem.utils.hashing import content_hash

        store = FAISSVectorStore(VectorStoreConfig(), dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "hello", "content_hash": content_hash("hello")})
        store.insert("id2", [0.4, 0.3, 0.2, 0.1], {"content": "world", "content_hash": content_hash("world")})

        all_items = store.list_all()
        assert len(all_items) == 2
        hashes = {m["content_hash"] for _, m in all_items}
        assert content_hash("hello") in hashes
        assert content_hash("world") in hashes

    def test_list_all_with_filters(self):
        from widemem.core.types import VectorStoreConfig
        from widemem.storage.vector.faiss_store import FAISSVectorStore

        store = FAISSVectorStore(VectorStoreConfig(), dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "a", "tier": "fact"})
        store.insert("id2", [0.4, 0.3, 0.2, 0.1], {"content": "b", "tier": "summary"})
        store.insert("id3", [0.2, 0.3, 0.4, 0.1], {"content": "c", "tier": "fact"})

        facts = store.list_all(filters={"tier": "fact"})
        assert len(facts) == 2
        summaries = store.list_all(filters={"tier": "summary"})
        assert len(summaries) == 1

    def test_list_all_no_filters(self):
        from widemem.core.types import VectorStoreConfig
        from widemem.storage.vector.faiss_store import FAISSVectorStore

        store = FAISSVectorStore(VectorStoreConfig(), dimensions=4)
        store.insert("id1", [0.1, 0.2, 0.3, 0.4], {"content": "a"})
        store.insert("id2", [0.4, 0.3, 0.2, 0.1], {"content": "b"})

        all_items = store.list_all()
        assert len(all_items) == 2

    def test_conflict_resolver_negative_fact_index(self):
        from widemem.conflict.batch_resolver import BatchConflictResolver
        from widemem.core.types import Fact, LLMConfig
        from widemem.providers.llm.base import BaseLLM

        class FakeLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig())
            def generate(self, prompt, system=None):
                return ""
            def generate_json(self, prompt, system=None):
                return {"actions": [
                    {"fact_index": -1, "action": "add"},
                ]}

        resolver = BatchConflictResolver(FakeLLM())
        facts = [Fact(content="only fact", importance=5.0)]
        existing = [MemorySearchResult(
            memory=Memory(content="old"),
            similarity_score=0.9,
        )]
        actions = resolver.resolve(facts, existing)
        assert len(actions) == 1
        assert actions[0].fact == "only fact"
        assert actions[0].action.value == "add"

    def test_conflict_resolver_duplicate_fact_index(self):
        from widemem.conflict.batch_resolver import BatchConflictResolver
        from widemem.core.types import Fact, LLMConfig
        from widemem.providers.llm.base import BaseLLM

        class FakeLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig())
            def generate(self, prompt, system=None):
                return ""
            def generate_json(self, prompt, system=None):
                return {"actions": [
                    {"fact_index": 0, "action": "add"},
                    {"fact_index": 0, "action": "add"},
                ]}

        resolver = BatchConflictResolver(FakeLLM())
        facts = [Fact(content="test", importance=5.0)]
        existing = [MemorySearchResult(
            memory=Memory(content="old"),
            similarity_score=0.9,
        )]
        actions = resolver.resolve(facts, existing)
        fact_actions = [a for a in actions if a.fact == "test"]
        assert len(fact_actions) == 1

    def test_conflict_resolver_missing_fact_index(self):
        from widemem.conflict.batch_resolver import BatchConflictResolver
        from widemem.core.types import Fact, LLMConfig
        from widemem.providers.llm.base import BaseLLM

        class FakeLLM(BaseLLM):
            def __init__(self):
                super().__init__(LLMConfig())
            def generate(self, prompt, system=None):
                return ""
            def generate_json(self, prompt, system=None):
                return {"actions": [
                    {"action": "add"},
                ]}

        resolver = BatchConflictResolver(FakeLLM())
        facts = [Fact(content="test", importance=5.0)]
        existing = [MemorySearchResult(
            memory=Memory(content="old"),
            similarity_score=0.9,
        )]
        actions = resolver.resolve(facts, existing)
        assert len(actions) == 1
        assert actions[0].fact == "test"
        assert actions[0].action.value == "add"



from __future__ import annotations

import threading
from typing import Callable, List, Optional

from widemem.conflict.batch_resolver import BatchConflictResolver
from widemem.core.types import (
    ActionItem,
    Fact,
    Memory,
    MemoryAction,
    MemorySearchResult,
    YMYLConfig,
)
from widemem.extraction.base import BaseExtractor
from widemem.providers.embeddings.base import BaseEmbedder
from widemem.retrieval.active import ActiveRetrieval, Clarification
from widemem.scoring.ymyl import is_ymyl_strong
from widemem.storage.history import HistoryStore
from widemem.storage.vector.base import BaseVectorStore
from widemem.utils.hashing import content_hash


class AddResult:
    __slots__ = ("memories", "clarifications")

    def __init__(
        self,
        memories: List[Memory],
        clarifications: Optional[List[Clarification]] = None,
    ) -> None:
        self.memories = memories
        self.clarifications = clarifications or []

    @property
    def has_clarifications(self) -> bool:
        return len(self.clarifications) > 0


class MemoryPipeline:
    def __init__(
        self,
        extractor: BaseExtractor,
        resolver: BatchConflictResolver,
        embedder: BaseEmbedder,
        vector_store: BaseVectorStore,
        history: HistoryStore,
        active_retrieval: Optional[ActiveRetrieval] = None,
        ymyl_active_retrieval: Optional[ActiveRetrieval] = None,
        ymyl_config: Optional[YMYLConfig] = None,
    ) -> None:
        self.extractor = extractor
        self.resolver = resolver
        self.embedder = embedder
        self.vector_store = vector_store
        self.history = history
        self.active_retrieval = active_retrieval
        self.ymyl_active_retrieval = ymyl_active_retrieval
        self.ymyl_config = ymyl_config or YMYLConfig()
        self._lock = threading.Lock()

    def process(
        self,
        text: str,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        on_clarification: Optional[Callable[[List[Clarification]], Optional[List[str]]]] = None,
    ) -> AddResult:
        facts = self.extractor.extract(text)
        if not facts:
            return AddResult(memories=[])

        with self._lock:
            existing = self._find_existing(facts, user_id=user_id, agent_id=agent_id)

            clarifications: List[Clarification] = []
            active = self.active_retrieval
            if not active and self.ymyl_active_retrieval and self.ymyl_config.enabled:
                has_ymyl_fact = any(
                    is_ymyl_strong(f.content, self.ymyl_config) for f in facts
                )
                if has_ymyl_fact:
                    active = self.ymyl_active_retrieval

            if active and existing:
                clarifications = active.detect_conflicts(facts, existing)

                if clarifications and on_clarification:
                    responses = on_clarification(clarifications)
                    if responses is None:
                        return AddResult(memories=[], clarifications=clarifications)

            actions = self.resolver.resolve(facts, existing)
            existing_hashes = {
                m.memory.metadata.get("content_hash") or content_hash(m.memory.content)
                for m in existing
            }
            memories = self._execute_actions(actions, user_id=user_id, agent_id=agent_id, run_id=run_id, existing_hashes=existing_hashes)
            return AddResult(memories=memories, clarifications=clarifications)

    def _find_existing(
        self,
        facts: list[Fact],
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
    ) -> list[MemorySearchResult]:
        all_results: dict[str, MemorySearchResult] = {}

        embeddings = self.embedder.embed_batch([f.content for f in facts])

        filters: dict[str, str] = {}
        if user_id:
            filters["user_id"] = user_id
        if agent_id:
            filters["agent_id"] = agent_id

        for embedding in embeddings:
            results = self.vector_store.search(
                vector=embedding,
                top_k=5,
                filters=filters or None,
            )
            for id, score, metadata in results:
                if id not in all_results:
                    all_results[id] = MemorySearchResult(
                        memory=Memory(
                            id=id,
                            content=metadata.get("content", ""),
                            user_id=metadata.get("user_id"),
                            agent_id=metadata.get("agent_id"),
                            importance=metadata.get("importance", 5.0),
                        ),
                        similarity_score=score,
                    )

        return list(all_results.values())

    def _hash_exists(self, hash_val: str, existing_hashes: set) -> bool:
        return hash_val in existing_hashes

    def _execute_actions(
        self,
        actions: list[ActionItem],
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        run_id: Optional[str] = None,
        existing_hashes: Optional[set] = None,
    ) -> list[Memory]:
        existing_hashes = existing_hashes or set()
        results = []

        for action in actions:
            if action.action == MemoryAction.ADD:
                new_hash = content_hash(action.fact)
                if self._hash_exists(new_hash, existing_hashes):
                    continue

                memory = Memory(
                    content=action.fact,
                    user_id=user_id,
                    agent_id=agent_id,
                    run_id=run_id,
                    importance=action.importance,
                    content_hash=new_hash,
                )
                embedding = self.embedder.embed(action.fact)
                self.vector_store.insert(
                    id=memory.id,
                    vector=embedding,
                    metadata=self._memory_to_metadata(memory),
                )
                existing_hashes.add(new_hash)
                self.history.log(memory.id, MemoryAction.ADD, new_content=action.fact)
                results.append(memory)

            elif action.action == MemoryAction.UPDATE and action.target_id:
                existing = self.vector_store.get(action.target_id)
                old_content = None
                if existing:
                    old_content = existing[1].get("content")

                new_hash = content_hash(action.fact)
                if existing and existing[1].get("content_hash") == new_hash:
                    continue

                memory = Memory(
                    id=action.target_id,
                    content=action.fact,
                    user_id=user_id,
                    agent_id=agent_id,
                    run_id=run_id,
                    importance=action.importance,
                    content_hash=new_hash,
                )
                embedding = self.embedder.embed(action.fact)
                self.vector_store.update(
                    id=action.target_id,
                    vector=embedding,
                    metadata=self._memory_to_metadata(memory),
                )
                self.history.log(
                    action.target_id, MemoryAction.UPDATE,
                    old_content=old_content, new_content=action.fact,
                )
                results.append(memory)

            elif action.action == MemoryAction.DELETE and action.target_id:
                existing = self.vector_store.get(action.target_id)
                old_content = existing[1].get("content") if existing else None
                self.vector_store.delete(action.target_id)
                self.history.log(
                    action.target_id, MemoryAction.DELETE,
                    old_content=old_content,
                )

        return results

    def _memory_to_metadata(self, memory: Memory) -> dict:
        return {
            "content": memory.content,
            "user_id": memory.user_id,
            "agent_id": memory.agent_id,
            "run_id": memory.run_id,
            "tier": memory.tier.value,
            "importance": memory.importance,
            "content_hash": memory.content_hash,
            "created_at": memory.created_at.isoformat(),
            "updated_at": memory.updated_at.isoformat(),
        }



from __future__ import annotations


def normalize_importance(importance: float) -> float:
    """Normalize importance from 0-10 scale to 0-1 scale."""
    return max(min(importance, 10.0), 0.0) / 10.0



from __future__ import annotations


class IDMapper:
    """Maps UUIDs to simple integers for LLM calls, preventing hallucinated IDs."""

    def __init__(self) -> None:
        self._uuid_to_int: dict[str, int] = {}
        self._int_to_uuid: dict[int, str] = {}
        self._next = 1

    def add(self, uuid: str) -> int:
        if uuid in self._uuid_to_int:
            return self._uuid_to_int[uuid]
        idx = self._next
        self._uuid_to_int[uuid] = idx
        self._int_to_uuid[idx] = uuid
        self._next += 1
        return idx

    def to_uuid(self, idx: int) -> str | None:
        return self._int_to_uuid.get(idx)

    def to_int(self, uuid: str) -> int | None:
        return self._uuid_to_int.get(uuid)



"""Temporal retrieval: time-weighted scoring and time range filters."""

from datetime import datetime, timedelta

from widemem import WideMemory, MemoryConfig
from widemem.core.types import DecayFunction, ScoringConfig

# Configure with exponential decay — recent memories score higher
config = MemoryConfig(
    scoring=ScoringConfig(
        decay_function=DecayFunction.EXPONENTIAL,
        decay_rate=0.02,  # Faster decay
        similarity_weight=0.4,
        importance_weight=0.3,
        recency_weight=0.3,  # Higher recency weight
    ),
)
mem = WideMemory(config=config)

mem.add("Started a new job at Google", user_id="alice")
mem.add("Learning Rust programming", user_id="alice")

# Search with time filter — only memories from last 7 days
results = mem.search(
    "what is alice doing",
    user_id="alice",
    time_after=datetime.utcnow() - timedelta(days=7),
)
for r in results:
    print(f"  [{r.final_score:.3f}] sim={r.similarity_score:.2f} "
          f"imp={r.importance_score:.2f} rec={r.temporal_score:.2f} "
          f"| {r.memory.content}")



from __future__ import annotations

from abc import ABC, abstractmethod

from widemem.core.types import EmbeddingConfig


class BaseEmbedder(ABC):
    def __init__(self, config: EmbeddingConfig) -> None:
        self.config = config

    @abstractmethod
    def embed(self, text: str) -> list[float]:
        ...

    @abstractmethod
    def embed_batch(self, texts: list[str]) -> list[list[float]]:
        ...

    @property
    def dimensions(self) -> int:
        return self.config.dimensions



from __future__ import annotations

import json

from openai import OpenAI

from widemem.core.exceptions import ProviderError
from widemem.core.types import LLMConfig
from widemem.providers.llm.base import BaseLLM


class OpenAILLM(BaseLLM):
    def __init__(self, config: LLMConfig) -> None:
        super().__init__(config)
        self.client = OpenAI(
            api_key=config.api_key,
            base_url=config.base_url,
        )

    def _generate(self, prompt: str, system: str | None = None) -> str:
        messages = []
        if system:
            messages.append({"role": "system", "content": system})
        messages.append({"role": "user", "content": prompt})

        response = self.client.chat.completions.create(
            model=self.config.model,
            messages=messages,
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens,
        )
        content = response.choices[0].message.content
        if content is None:
            raise ProviderError("LLM returned empty response")
        return content

    def _generate_json(self, prompt: str, system: str | None = None) -> dict:
        messages = []
        if system:
            messages.append({"role": "system", "content": system})
        messages.append({"role": "user", "content": prompt})

        response = self.client.chat.completions.create(
            model=self.config.model,
            messages=messages,
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens,
            response_format={"type": "json_object"},
        )
        content = response.choices[0].message.content
        if content is None:
            raise ProviderError("LLM returned empty response")
        try:
            return json.loads(content)
        except json.JSONDecodeError as e:
            raise ProviderError(f"LLM returned invalid JSON: {e}") from e



CONTRADICTION_DETECTION_SYSTEM = (
    "You are a memory contradiction detector. Given a new fact and existing memories, "
    "determine if any contradictions or ambiguities exist.\n\n"
    "Rules:\n"
    "- A contradiction exists when new information directly conflicts with existing memory "
    '(e.g., "lives in Berlin" vs "lives in Paris")\n'
    "- An ambiguity exists when the relationship between new and existing is unclear "
    '(e.g., "in Berlin" could mean moved or visiting)\n'
    "- Temporal changes are common ambiguities: people move, change jobs, etc.\n"
    "- Only flag genuine contradictions/ambiguities, not additions of new information\n\n"
    "Respond with JSON only."
)

CONTRADICTION_DETECTION_PROMPT = (
    "New fact:\n{new_fact}\n\n"
    "Existing memories:\n{existing_memories}\n\n"
    "Are there contradictions or ambiguities? Respond with:\n"
    '{{\"has_conflict\": true/false, \"conflicts\": ['
    '{{\"new_fact\": \"the new fact\", '
    '\"existing_memory_id\": 1, '
    '\"existing_content\": \"the existing memory\", '
    '\"type\": \"contradiction\" or \"ambiguity\", '
    '\"question\": \"clarifying question to resolve this\"}}]}}\n\n'
    'If no conflicts: {{\"has_conflict\": false, \"conflicts\": []}}'
)
